from collections import OrderedDict
from functools import lru_cache
from pprint import pprint

from py42.exceptions import Py42ForbiddenError, Py42BadRequestError


from code42cli.errors import (
    UserAlreadyAddedError,
    UserNotInLegalHoldError,
    LegalHoldNotFoundOrPermissionDeniedError,
)
from code42cli.util import (
    format_to_table,
    find_format_width,
    format_string_list_to_columns,
    get_user_id,
)
from code42cli.bulk import run_bulk_process
from code42cli.file_readers import create_csv_reader
from code42cli.logger import get_main_cli_logger

_MATTER_KEYS_MAP = OrderedDict()
_MATTER_KEYS_MAP[u"legalHoldUid"] = u"Matter ID"
_MATTER_KEYS_MAP[u"name"] = u"Name"
_MATTER_KEYS_MAP[u"description"] = u"Description"
_MATTER_KEYS_MAP[u"creator_username"] = u"Creator"
_MATTER_KEYS_MAP[u"creationDate"] = u"Creation Date"

logger = get_main_cli_logger()


def add_user(sdk, matter_id, username):
    user_id = get_user_id(sdk, username)
    matter = _check_matter_is_accessible(sdk, matter_id)
    try:
        sdk.legalhold.add_to_matter(user_id, matter_id)
    except Py42BadRequestError as e:
        if u"USER_ALREADY_IN_HOLD" in e.response.text:
            matter_id_and_name_text = u"legal hold matter id={}, name={}".format(
                matter_id, matter[u"name"]
            )
            raise UserAlreadyAddedError(username, matter_id_and_name_text)
        raise


def remove_user(sdk, matter_id, username):
    _check_matter_is_accessible(sdk, matter_id)
    membership_id = _get_legal_hold_membership_id_for_user_and_matter(sdk, username, matter_id)
    sdk.legalhold.remove_from_matter(membership_id)


def get_matters(sdk):
    matters = _get_all_active_matters(sdk)
    if matters:
        rows, column_size = find_format_width(matters, _MATTER_KEYS_MAP)
        format_to_table(rows, column_size)


def add_bulk_users(sdk, file_name):
    reader = create_csv_reader(file_name)
    run_bulk_process(
        lambda matter_id, username: add_user(sdk, matter_id, username), reader,
    )


def remove_bulk_users(sdk, file_name):
    reader = create_csv_reader(file_name)
    run_bulk_process(
        lambda matter_id, username: remove_user(sdk, matter_id, username), reader,
    )


def show_matter(sdk, matter_id, include_inactive=False, include_policy=False):
    matter = _check_matter_is_accessible(sdk, matter_id)
    matter[u"creator_username"] = matter[u"creator"][u"username"]

    # if `active` is None then all matters (whether active or inactive) are returned. True returns
    # only those that are active.
    active = None if include_inactive else True
    memberships = _get_legal_hold_memberships_for_matter(sdk, matter_id, active=active)
    active_usernames = [member[u"user"][u"username"] for member in memberships if member[u"active"]]
    inactive_usernames = [
        member[u"user"][u"username"] for member in memberships if not member[u"active"]
    ]

    rows, column_size = find_format_width([matter], _MATTER_KEYS_MAP)

    print(u"")
    format_to_table(rows, column_size)
    if active_usernames:
        print(u"\nActive matter members:\n")
        format_string_list_to_columns(active_usernames)
    else:
        print("\nNo active matter members.\n")

    if include_inactive:
        if inactive_usernames:
            print(u"\nInactive matter members:\n")
            format_string_list_to_columns(inactive_usernames)
        else:
            print("No inactive matter members.\n")

    if include_policy:
        _get_and_print_preservation_policy(sdk, matter[u"holdPolicyUid"])
        print(u"")


def _get_and_print_preservation_policy(sdk, policy_uid):
    preservation_policy = sdk.legalhold.get_policy_by_uid(policy_uid)
    print(u"\nPreservation Policy:\n")
    pprint(preservation_policy._data_root)


def _get_legal_hold_membership_id_for_user_and_matter(sdk, username, matter_id):
    user_id = get_user_id(sdk, username)
    memberships = _get_legal_hold_memberships_for_matter(sdk, matter_id, active=True)
    for member in memberships:
        if member[u"user"][u"userUid"] == user_id:
            return member[u"legalHoldMembershipUid"]
    raise UserNotInLegalHoldError(username, matter_id)


def _get_legal_hold_memberships_for_matter(sdk, matter_id, active=True):
    memberships_generator = sdk.legalhold.get_all_matter_custodians(
        legal_hold_uid=matter_id, active=active
    )
    memberships = [
        member for page in memberships_generator for member in page[u"legalHoldMemberships"]
    ]
    return memberships


def _get_all_active_matters(sdk):
    matters_generator = sdk.legalhold.get_all_matters()
    matters = [
        matter for page in matters_generator for matter in page[u"legalHolds"] if matter[u"active"]
    ]
    for matter in matters:
        matter[u"creator_username"] = matter[u"creator"][u"username"]
    return matters


@lru_cache(maxsize=None)
def _check_matter_is_accessible(sdk, matter_id):
    try:
        matter = sdk.legalhold.get_matter_by_uid(matter_id)
        return matter
    except (Py42BadRequestError, Py42ForbiddenError):
        raise LegalHoldNotFoundOrPermissionDeniedError(matter_id)



import os, inspect

from code42cli.compat import open, str
from code42cli.worker import Worker
from code42cli.logger import get_main_cli_logger
from code42cli.args import SDK_ARG_NAME, PROFILE_ARG_NAME
from code42cli.progress_bar import ProgressBar


_logger = get_main_cli_logger()


class BulkCommandType(object):
    ADD = u"add"
    REMOVE = u"remove"

    def __iter__(self):
        return iter([self.ADD, self.REMOVE])


def generate_template(handler, path=None):
    """Looks at the parameter names of `handler` and creates a file with the same column names. If 
    `handler` only has one parameter that is not `sdk` or `profile`, it will create a blank file. 
    This is useful for commands such as `remove` which only require a list of users.
    """
    path = path or os.path.join(os.getcwd(), u"{}.csv".format(str(handler.__name__)))
    args = [
        arg
        for arg in inspect.getargspec(handler).args
        if arg != SDK_ARG_NAME and arg != PROFILE_ARG_NAME
    ]

    if len(args) <= 1:
        _logger.print_info(
            u"A blank file was generated because there are no csv headers needed for this command. "
            u"Simply enter one {} per line.".format(args[0])
        )
        # Set args to None so that we don't make a header out of the single arg.
        args = None

    _write_template_file(path, args)


def _write_template_file(path, columns=None):
    with open(path, u"w", encoding=u"utf8") as new_file:
        if columns:
            new_file.write(u",".join(columns))


def run_bulk_process(row_handler, reader):
    """Runs a bulk process.
    
    Args: 
        row_handler (callable): A callable that you define to process values from the row as 
            either *args or **kwargs.
        reader: (CSVReader or FlatFileReader, optional): A generator that reads rows and yields data into 
            `row_handler`. If None, it will use a CSVReader. Defaults to None.
    """
    processor = _create_bulk_processor(row_handler, reader)
    processor.run()


def _create_bulk_processor(row_handler, reader):
    """A factory method to create the bulk processor, useful for testing purposes."""
    return BulkProcessor(row_handler, reader)


class BulkProcessor(object):
    """A class for bulk processing a file. 
    
    Args:
        row_handler (callable): A callable that you define to process values from the row as 
            either *args or **kwargs. For example, if it's a csv file with header `prop_a,prop_b` 
            and first row `1,test`, then `row_handler` should receive kwargs 
            `prop_a: '1', prop_b: 'test'` when processing the first row. If it's a flat file, then 
            `row_handler` only needs to take an extra arg.
        reader (CSVReader or FlatFileReader): A generator that reads rows and yields data into `row_handler`.
    """

    def __init__(self, row_handler, reader, worker=None, progress_bar=None):
        total = reader.get_rows_count()
        self.file_path = reader.file_path
        self._row_handler = row_handler
        self._reader = reader
        self.__worker = worker or Worker(5, total)
        self._stats = self.__worker.stats
        self._progress_bar = progress_bar or ProgressBar(total)

    def run(self):
        """Processes the csv file specified in the ctor, calling `self.row_handler` on each row."""
        with open(self.file_path, newline=u"", encoding=u"utf8") as bulk_file:
            for row in self._reader(bulk_file=bulk_file):
                self._process_row(row)
            self.__worker.wait()
        self._print_results()

    def _process_row(self, row):
        if isinstance(row, dict):
            self._process_csv_row(row)
        elif row:
            self._process_flat_file_row(row.strip())

    def _process_csv_row(self, row):
        # Removes problems from including extra comments. Error messages from out of order args
        # are more indicative this way too.
        row.pop(None, None)
        row_values = {key: val if val != u"" else None for key, val in row.items()}
        self.__worker.do_async(
            lambda *args, **kwargs: self._handle_row(*args, **kwargs), **row_values
        )

    def _process_flat_file_row(self, row):
        if row:
            self.__worker.do_async(lambda *args, **kwargs: self._handle_row(*args, **kwargs), row)

    def _handle_row(self, *args, **kwargs):
        message = str(self._stats)
        self._progress_bar.update(self._stats.total_processed, message)
        self._row_handler(*args, **kwargs)

    def _print_results(self):
        self._progress_bar.clear_bar_and_print_final(str(self._stats))
        if self._stats.total_errors:
            logger = get_main_cli_logger()
            logger.print_errors_occurred_message()



from collections import OrderedDict

from py42.exceptions import Py42InternalServerError
from py42.util import format_json


from code42cli.errors import InvalidRuleTypeError
from code42cli.util import format_to_table, find_format_width, get_user_id
from code42cli.bulk import run_bulk_process
from code42cli.file_readers import create_csv_reader
from code42cli.logger import get_main_cli_logger
from code42cli.cmds.alerts.rules.enums import AlertRuleTypes


_HEADER_KEYS_MAP = OrderedDict()
_HEADER_KEYS_MAP[u"observerRuleId"] = u"RuleId"
_HEADER_KEYS_MAP[u"name"] = u"Name"
_HEADER_KEYS_MAP[u"severity"] = u"Severity"
_HEADER_KEYS_MAP[u"type"] = u"Type"
_HEADER_KEYS_MAP[u"ruleSource"] = u"Source"
_HEADER_KEYS_MAP[u"isEnabled"] = u"Enabled"


def add_user(sdk, profile, rule_id, username):
    user_id = get_user_id(sdk, username)
    rules = _get_rule_metadata(sdk, rule_id)
    try:
        if rules:
            sdk.alerts.rules.add_user(rule_id, user_id)
    except Py42InternalServerError as e:
        _check_if_system_rule(sdk, rules)
        raise


def remove_user(sdk, profile, rule_id, username):
    user_id = get_user_id(sdk, username)
    rules = _get_rule_metadata(sdk, rule_id)
    try:
        if rules:
            sdk.alerts.rules.remove_user(rule_id, user_id)
    except Py42InternalServerError as e:
        _check_if_system_rule(sdk, rules)
        raise


def _get_all_rules_metadata(sdk):
    rules_generator = sdk.alerts.rules.get_all()
    selected_rules = [rule for rules in rules_generator for rule in rules[u"ruleMetadata"]]
    return _handle_rules_results(sdk, selected_rules)


def _get_rule_metadata(sdk, rule_id):
    rules = sdk.alerts.rules.get_by_observer_id(rule_id)[u"ruleMetadata"]
    return _handle_rules_results(sdk, rules, rule_id)


def _handle_rules_results(sdk, rules, rule_id=None):
    id_msg = u"with RuleId {} ".format(rule_id) if rule_id else u""
    msg = u"No alert rules {0}found.".format(id_msg)
    if not rules:
        get_main_cli_logger().print_and_log_info(msg)
    return rules


def _check_if_system_rule(sdk, rules):
    if rules and rules[0][u"isSystem"]:
        raise InvalidRuleTypeError(rules[0][u"observerRuleId"], rules[0][u"ruleSource"])


def get_rules(sdk, profile):
    selected_rules = _get_all_rules_metadata(sdk)
    if selected_rules:
        rows, column_size = find_format_width(selected_rules, _HEADER_KEYS_MAP)
        format_to_table(rows, column_size)


def add_bulk_users(sdk, profile, file_name):
    reader = create_csv_reader(file_name)
    run_bulk_process(lambda rule_id, username: add_user(sdk, profile, rule_id, username), reader)


def remove_bulk_users(sdk, profile, file_name):
    reader = create_csv_reader(file_name)
    run_bulk_process(lambda rule_id, username: remove_user(sdk, profile, rule_id, username), reader)


def show_rule(sdk, profile, rule_id):
    selected_rule = _get_rule_metadata(sdk, rule_id)
    rule_detail = None
    if selected_rule:
        rule_type = selected_rule[0][u"type"]
        if rule_type == AlertRuleTypes.EXFILTRATION:
            rule_detail = sdk.alerts.rules.exfiltration.get(rule_id)
        elif rule_type == AlertRuleTypes.CLOUD_SHARE:
            rule_detail = sdk.alerts.rules.cloudshare.get(rule_id)
        elif rule_type == AlertRuleTypes.FILE_TYPE_MISMATCH:
            rule_detail = sdk.alerts.rules.filetypemismatch.get(rule_id)
    if rule_detail:
        logger = get_main_cli_logger()
        logger.print_info(format_json(rule_detail.text))



import py42.sdk
import py42.settings.debug as debug

from code42cli.logger import get_main_cli_logger


def create_sdk(profile, is_debug_mode):
    if is_debug_mode:
        py42.settings.debug.level = debug.DEBUG
    try:
        password = profile.get_password()
        return py42.sdk.from_local_account(profile.authority_url, profile.username, password)
    except Exception:
        logger = get_main_cli_logger()
        logger.print_and_log_error(
            u"Invalid credentials or host address. "
            u"Verify your profile is set up correctly and that you are supplying the correct password."
        )
        exit(1)


def validate_connection(authority_url, username, password):
    try:
        py42.sdk.from_local_account(authority_url, username, password)
        return True
    except:
        return False



from code42cli.cmds.detectionlists import (
    DetectionList,
    DetectionListHandlers,
    load_user_descriptions,
    update_user,
    try_handle_user_already_added_error,
    DetectionListSubcommandLoader,
)
from code42cli.util import get_user_id
from code42cli.cmds.detectionlists.enums import DetectionLists

from py42.exceptions import Py42BadRequestError


class DepartingEmployeeSubcommandLoader(DetectionListSubcommandLoader):
    def __init__(self, root_command_name):
        super(DepartingEmployeeSubcommandLoader, self).__init__(root_command_name)
        handlers = _create_handlers()
        self.detection_list = DetectionList.create_departing_employee_list(handlers)
        self._cmd_loader = self.detection_list.subcommand_loader

    def load_commands(self):
        return self.detection_list.load_subcommands()


def _create_handlers():
    return DetectionListHandlers(
        add=add_departing_employee, remove=remove_departing_employee, load_add=_load_add_description
    )


def add_departing_employee(
    sdk, profile, username, cloud_alias=None, departure_date=None, notes=None
):
    """Adds an employee to the departing employee detection list.

    Args:
        sdk (py42.sdk.SDKClient): py42.
        profile (C42Profile): Your code42 profile.
        username (str): The username of the employee to add.
        cloud_alias (str): An alternative email address for another cloud service.
        departure_date (str): The date the employee is departing in format `yyyy-MM-dd`.
        notes: (str): Notes about the employee.
    """
    user_id = get_user_id(sdk, username)

    try:
        sdk.detectionlists.departing_employee.add(user_id, departure_date)
        update_user(sdk, user_id, cloud_alias, notes=notes)
    except Py42BadRequestError as err:
        list_name = u"{} list".format(DetectionLists.DEPARTING_EMPLOYEE)
        try_handle_user_already_added_error(err, username, list_name)
        raise


def remove_departing_employee(sdk, profile, username):
    user_id = get_user_id(sdk, username)
    sdk.detectionlists.departing_employee.remove(user_id)


def _load_add_description(argument_collection):
    load_user_descriptions(argument_collection)
    departure_date = argument_collection.arg_configs[u"departure_date"]
    departure_date.set_help(u"The date the employee is departing in format yyyy-MM-dd.")



from getpass import getpass

import keyring

from code42cli import PRODUCT_NAME
from code42cli.util import does_user_agree


def get_stored_password(profile):
    """Gets your currently stored password for the given profile name."""
    service_name = _get_keyring_service_name(profile.name)
    return keyring.get_password(service_name, profile.username)


def get_password_from_prompt():
    """Prompts you and returns what you input."""
    return getpass()


def set_password(profile, new_password):
    """Sets your password for the given profile name."""
    service_name = _get_keyring_service_name(profile.name)
    uses_file_storage = keyring.get_keyring().priority < 1
    if uses_file_storage and not _prompt_for_alternative_store():
        return

    keyring.set_password(service_name, profile.username, new_password)


def delete_password(profile):
    """Deletes password for the given profile name."""
    service_name = _get_keyring_service_name(profile.name)
    keyring.delete_password(service_name, profile.username)


def _get_keyring_service_name(profile_name):
    return u"{}::{}".format(PRODUCT_NAME, profile_name)


def _prompt_for_alternative_store():
    prompt = u"keyring is unavailable. Would you like to store in secure flat file? (y/n): "
    return does_user_agree(prompt)



class DetectionLists(object):
    DEPARTING_EMPLOYEE = u"departing-employee"
    HIGH_RISK_EMPLOYEE = u"high-risk-employee"


class DetectionListUserKeys(object):
    CLOUD_ALIAS = u"cloud_alias"
    USERNAME = u"username"
    NOTES = u"notes"
    RISK_TAG = u"risk_tag"


class RiskTags(object):
    FLIGHT_RISK = u"FLIGHT_RISK"
    HIGH_IMPACT_EMPLOYEE = u"HIGH_IMPACT_EMPLOYEE"
    ELEVATED_ACCESS_PRIVILEGES = u"ELEVATED_ACCESS_PRIVILEGES"
    PERFORMANCE_CONCERNS = u"PERFORMANCE_CONCERNS"
    SUSPICIOUS_SYSTEM_ACTIVITY = u"SUSPICIOUS_SYSTEM_ACTIVITY"
    POOR_SECURITY_PRACTICES = u"POOR_SECURITY_PRACTICES"
    CONTRACT_EMPLOYEE = u"CONTRACT_EMPLOYEE"

    def __iter__(self):
        return iter(
            [
                self.FLIGHT_RISK,
                self.HIGH_IMPACT_EMPLOYEE,
                self.ELEVATED_ACCESS_PRIVILEGES,
                self.PERFORMANCE_CONCERNS,
                self.SUSPICIOUS_SYSTEM_ACTIVITY,
                self.POOR_SECURITY_PRACTICES,
                self.CONTRACT_EMPLOYEE,
            ]
        )



from collections import OrderedDict
import inspect


PROFILE_HELP = u"The name of the Code42 CLI profile to use when executing this command."
SDK_ARG_NAME = u"sdk"
PROFILE_ARG_NAME = u"profile"


class ArgConfig(object):
    """Stores a set of argparse commands for later use by a command."""

    def __init__(self, *args, **kwargs):
        self._settings = {
            u"action": kwargs.get(u"action"),
            u"choices": kwargs.get(u"choices"),
            u"default": kwargs.get(u"default"),
            u"help": kwargs.get(u"help"),
            u"options_list": list(args),
            u"nargs": kwargs.get(u"nargs"),
            u"metavar": kwargs.get(u"metavar"),
            u"required": kwargs.get(u"required"),
        }

    @property
    def settings(self):
        return self._settings

    def set_choices(self, choices):
        self._settings[u"choices"] = choices

    def set_help(self, help):
        self._settings[u"help"] = help

    def add_short_option_name(self, short_name):
        self._settings[u"options_list"].append(short_name)

    def as_multi_val_param(self, nargs=u"+"):
        self._settings[u"nargs"] = nargs

    def set_required(self, required):
        self._settings[u"required"] = required


class ArgConfigCollection(object):
    def __init__(self):
        self._arg_configs = OrderedDict()

    @property
    def arg_configs(self):
        return self._arg_configs

    def append(self, name, arg_config):
        self._arg_configs[name] = arg_config

    def extend(self, arg_config_dict):
        self.arg_configs.update(arg_config_dict)


def get_auto_arg_configs(handler):
    """Looks at the parameter names of `handler` and builds an `ArgConfigCollection` containing
    `argparse` parameters based on them."""
    arg_configs = ArgConfigCollection()
    excluded_args = [SDK_ARG_NAME, u"profile", u"args", u"kwargs", u"self"]
    if callable(handler):
        argspec = inspect.getargspec(handler)
        filtered_argspec = {
            key: position for position, key in enumerate(argspec.args) if key not in excluded_args
        }
        num_optional_args = len(argspec.defaults) if argspec.defaults else 0
        num_positional_args = len(argspec.args) - num_optional_args
        num_required_cli_args = len(filtered_argspec) - num_optional_args

        for key in filtered_argspec:
            arg_config = _create_auto_args_config(
                key, filtered_argspec[key], num_positional_args, num_required_cli_args, argspec
            )
            _set_smart_defaults(arg_config)
            arg_configs.append(key, arg_config)

        if SDK_ARG_NAME in argspec.args:
            _build_sdk_arg_configs(arg_configs)

    return arg_configs


def _create_auto_args_config(key, position, num_positional_args, num_required_cli_args, argspec):
    default = None
    required = None
    param_name = key.replace(u"_", u"-")
    last_positional_arg_idx = num_positional_args - 1
    option_names = [u"--{}".format(param_name)]
    # positional arguments will come first, so if the arg position
    # is greater than the index of the last positional arg, it's a kwarg.
    if position > last_positional_arg_idx:
        # this is a keyword arg, treat it as an optional cli arg.
        default_value = argspec.defaults[position - num_positional_args]
        default = default_value
    elif num_required_cli_args > 1:
        # this is a positional arg, treat it as a required cli arg.
        required = True
    else:
        option_names = [param_name]
    return ArgConfig(*option_names, default=default, required=required)


def _set_smart_defaults(arg_config):
    default = arg_config.settings.get(u"default")
    # make the param not require a value (e.g. --enable) if the default value of
    # the param is a bool.
    if type(default) == bool:
        arg_config.settings[u"action"] = u"store_{}".format(default).lower()


def _build_sdk_arg_configs(arg_config_collection):
    """Add extra cli parameters that will always be relevant when a handler needs the sdk."""
    profile = ArgConfig(u"--profile", help=PROFILE_HELP)
    debug = ArgConfig(u"-d", u"--debug", action=u"store_true", help=u"Turn on Debug logging.")
    extras = {PROFILE_ARG_NAME: profile, u"debug": debug}
    arg_config_collection.extend(extras)



from __future__ import with_statement

import sqlite3
import os

from code42cli.util import get_user_project_path


class BaseCursorStore(object):
    _PRIMARY_KEY_COLUMN_NAME = u"cursor_id"
    _timestamp_column_name = u"OVERRIDE"
    _primary_key = u"OVERRIDE"

    def __init__(self, db_table_name, db_file_path=None):
        self._table_name = db_table_name
        if db_file_path is None:
            db_path = get_user_project_path(u"db")
            db_file = u"file_event_checkpoints.db"
            db_file_path = os.path.join(db_path, db_file)

        self._connection = sqlite3.connect(db_file_path)
        if self._is_empty():
            self._init_table()

    def _get(self, columns, primary_key):
        query = u"SELECT {0} FROM {1} WHERE {2}=?"
        query = query.format(columns, self._table_name, self._PRIMARY_KEY_COLUMN_NAME)
        with self._connection as conn:
            cursor = conn.cursor()
            cursor.execute(query, (primary_key,))
            return cursor.fetchall()

    def _set(self, column_name, new_value, primary_key):
        query = u"UPDATE {0} SET {1}=? WHERE {2}=?".format(
            self._table_name, column_name, self._PRIMARY_KEY_COLUMN_NAME
        )
        with self._connection as conn:
            conn.execute(query, (new_value, primary_key))

    def _delete(self, primary_key):
        query = u"DELETE FROM {0} WHERE {1}=?".format(
            self._table_name, self._PRIMARY_KEY_COLUMN_NAME
        )
        with self._connection as conn:
            conn.execute(query, (primary_key,))

    def _row_exists(self, primary_key):
        query = u"SELECT * FROM {0} WHERE {1}=?"
        query = query.format(self._table_name, self._PRIMARY_KEY_COLUMN_NAME)
        with self._connection as conn:
            cursor = conn.cursor()
            cursor.execute(query, (primary_key,))
            query_result = cursor.fetchone()
            if not query_result:
                return False
            return True

    def _drop_table(self):
        drop_query = u"DROP TABLE {0}".format(self._table_name)
        with self._connection as conn:
            conn.execute(drop_query)

    def _is_empty(self):
        table_count_query = u"""
            SELECT COUNT(name)
            FROM sqlite_master
            WHERE type='table' AND name=?
        """
        with self._connection as conn:
            cursor = conn.cursor()
            cursor.execute(table_count_query, (self._table_name,))
            query_result = cursor.fetchone()
            if query_result:
                return int(query_result[0]) <= 0

    def _init_table(self):
        columns = u"{0}, {1}".format(self._PRIMARY_KEY_COLUMN_NAME, self._timestamp_column_name)
        create_table_query = u"CREATE TABLE {0} ({1})".format(self._table_name, columns)
        with self._connection as conn:
            conn.execute(create_table_query)

    def _insert_new_row(self):
        insert_query = u"INSERT INTO {0} VALUES(?, null)".format(self._table_name)
        with self._connection as conn:
            conn.execute(insert_query, (self._primary_key,))

    def get_stored_cursor_timestamp(self):
        """Gets the last stored date observed timestamp."""
        rows = self._get(self._timestamp_column_name, self._primary_key)
        if rows and rows[0]:
            return rows[0][0]

    def replace_stored_cursor_timestamp(self, new_date_observed_timestamp):
        """Replaces the last stored date observed timestamp with the given one."""
        self._set(
            column_name=self._timestamp_column_name,
            new_value=new_date_observed_timestamp,
            primary_key=self._primary_key,
        )

    def clean(self):
        """Removes profile cursor data from store."""
        self._delete(self._primary_key)


class FileEventCursorStore(BaseCursorStore):
    _timestamp_column_name = u"insertionTimestamp"

    def __init__(self, profile_name, db_file_path=None):
        self._primary_key = profile_name
        super(FileEventCursorStore, self).__init__(u"file_event_checkpoints", db_file_path)
        if not self._row_exists(self._primary_key):
            self._insert_new_row()


class AlertCursorStore(BaseCursorStore):
    _timestamp_column_name = u"createdAt"

    def __init__(self, profile_name, db_file_path=None):
        self._primary_key = profile_name
        super(AlertCursorStore, self).__init__(u"alert_checkpoints", db_file_path)
        if not self._row_exists(self._primary_key):
            self._insert_new_row()


def get_file_event_cursor_store(profile_name):
    return FileEventCursorStore(profile_name)


def get_alert_cursor_store(profile_name):
    return AlertCursorStore(profile_name)


def get_all_cursor_stores_for_profile(profile_name):
    return [FileEventCursorStore(profile_name), AlertCursorStore(profile_name)]



import inspect

from code42cli import profile as cliprofile
from code42cli.args import get_auto_arg_configs, SDK_ARG_NAME, PROFILE_ARG_NAME
from code42cli.sdk_client import create_sdk


class DictObject(object):
    def __init__(self, _dict):
        self.__dict__ = _dict


class Command(object):
    """Represents a function that a CLI user can execute. Add a command to
    `code42cli.main._load_top_commands` or as a subcommand of one those
    commands to make it available for use.

    Args:
        name (str or unicode): The name of the command. For example, in
            `code42 profile show`, "show" is the name, while "profile"
            is the name of the parent command.

        description (str or unicode): Descriptive text to be displayed when using -h.

        usage (str or unicode, optional): A usage example to be displayed when using -h.
        handler (function, optional): The function to be exectued when the command is run.

        arg_customizer (function, optional): A function accepting a single `ArgCollection`
            parameter that allows for editing the collection when `get_arg_configs` is run.

        subcommand_loader (SubcommandLoader, optional): An object that can load subcommands 
            for this command.

        use_single_arg_obj (bool, optional): When True, causes all parameters sent to
            `__call__` to be consolidated in an object with attribute names dictated
            by the parameter names. That object is passed to `handler`'s `arg` parameter.
    """

    def __init__(
        self,
        name,
        description,
        usage=None,
        handler=None,
        arg_customizer=None,
        subcommand_loader=None,
        use_single_arg_obj=None,
    ):
        self._name = name
        self._description = description
        self._usage = usage
        self._handler = handler
        self._arg_customizer = arg_customizer
        self._subcommand_loader = subcommand_loader
        self._use_single_arg_obj = use_single_arg_obj
        self._subcommands = []

    def __call__(self, *args, **kwargs):
        """Passes the parsed argparse args to the handler, or
        shows the help of for this command if there is no handler
        (common in commands that are simply groups of subcommands).
        """
        if callable(self._handler):
            kvps = _get_arg_kvps(args[0], self._handler)
            if self._use_single_arg_obj:
                kvps = _kvps_to_obj(kvps)
            return self._handler(**kvps)
        help_func = kwargs.pop(u"help_func", None)
        if help_func:
            return help_func()

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        return self._description

    @property
    def usage(self):
        return self._usage

    @property
    def subcommands(self):
        return self._subcommands

    @property
    def subcommand_loader(self):
        return self._subcommand_loader

    def load_subcommands(self):
        self._subcommands = (
            self._subcommand_loader.load_commands() if self._subcommand_loader else []
        )
        return self._subcommands

    def get_arg_configs(self):
        """Returns a collection of argparse configurations based on
        the parameter names of `handler` and any user customizations."""
        arg_config_collection = get_auto_arg_configs(self._handler)
        if callable(self._arg_customizer):
            self._arg_customizer(arg_config_collection)

        return arg_config_collection.arg_configs


def _get_arg_kvps(parsed_args, handler):
    # transform parsed args from `argparse` into a dict
    parsed_args = vars(parsed_args)
    kvps = {key.replace(u"-", u"_"): val for (key, val) in parsed_args.items()}
    kvps.pop(u"func", None)
    return _inject_params(kvps, handler)


def _inject_params(kvps, handler):
    """Automatically populates parameters named "sdk" or "profile" with instances of the sdk and 
    profile, respectively.

    Args:
        kvps (dict): A dictionary of the parsed command line arguments.
        handler (callable): The function or command responsible for processing the command line 
            arguments.

    Returns:
        dict: The dictionary of parsed command line arguments with possibly additional populated 
            fields.
    """
    if _handler_has_arg(SDK_ARG_NAME, handler):
        profile_name = kvps.pop(PROFILE_ARG_NAME, None)
        debug = kvps.pop(u"debug", None)

        profile = cliprofile.get_profile(profile_name)
        kvps[SDK_ARG_NAME] = create_sdk(profile, debug)

        if _handler_has_arg(PROFILE_ARG_NAME, handler):
            kvps[PROFILE_ARG_NAME] = profile
    return kvps


def _handler_has_arg(arg_name, handler):
    argspec = inspect.getargspec(handler)
    return arg_name in argspec.args


def _kvps_to_obj(kvps):
    new_kvps = {key: kvps[key] for key in kvps if key in [SDK_ARG_NAME, PROFILE_ARG_NAME]}
    new_kvps[u"args"] = DictObject(kvps)
    return new_kvps


class SubcommandLoader(object):
    """Responsible for creating subcommands for it's root command. It is also useful for getting 
    command information ahead of time, as in the example of tab completion."""

    def __init__(self, root_command_name):
        self.root = root_command_name

    @property
    def names(self):
        """The names of all the subcommands in this subcommabd loader's root command."""
        sub_cmds = self.load_commands()
        return [cmd.name for cmd in sub_cmds]

    @property
    def subtrees(self):
        """All subcommands for this subcommand loader's root command mapped to their given 
        subcommand loaders."""
        cmds = self.load_commands()
        results = {}
        for cmd in cmds:
            subcommand_loader = cmd.subcommand_loader
            if subcommand_loader:
                results[cmd.name] = subcommand_loader
        return results

    def load_commands(self):
        return []



"""
This module handles import compatibility issues between Python 2 and
Python 3.
"""
# pylint: disable=undefined-variable,import-error,unused-import,no-name-in-module

import sys

_ver = sys.version_info

#: Python 2.x?
is_py2 = _ver[0] == 2

if is_py2:
    from urlparse import urljoin, urlparse

    def _str(obj):
        return unicode(obj)

    str = _str

    import io

    open = io.open

    import repr as reprlib

    import Queue as queue

    range = xrange
else:
    from urllib.parse import urljoin, urlparse

    str = str
    open = open

    import reprlib

    import queue

    range = range



import os

from configparser import ConfigParser

import code42cli.util as util
from code42cli.compat import str
from code42cli.logger import get_main_cli_logger


class NoConfigProfileError(Exception):
    def __init__(self, profile_arg_name=None):
        message = (
            u"Profile '{}' does not exist.".format(profile_arg_name)
            if profile_arg_name
            else u"Profile does not exist."
        )
        super(NoConfigProfileError, self).__init__(message)


class ConfigAccessor(object):
    DEFAULT_VALUE = u"__DEFAULT__"
    AUTHORITY_KEY = u"c42_authority_url"
    USERNAME_KEY = u"c42_username"
    IGNORE_SSL_ERRORS_KEY = u"ignore-ssl-errors"
    DEFAULT_PROFILE = u"default_profile"
    _INTERNAL_SECTION = u"Internal"

    def __init__(self, parser):
        self.parser = parser
        file_name = u"config.cfg"
        self.path = os.path.join(util.get_user_project_path(), file_name)
        if not os.path.exists(self.path):
            self._create_internal_section()
            self._save()
        else:
            self.parser.read(self.path)

    def get_profile(self, name=None):
        """Returns the profile with the given name.
        If name is None, returns the default profile.
        If the name does not exist or there is no existing profile, it will throw an exception.
        """
        name = name or self._default_profile_name
        if name not in self._get_sections() or name == self.DEFAULT_VALUE:
            name = name if name != self.DEFAULT_VALUE else None
            raise NoConfigProfileError(name)
        return self._get_profile(name)

    def get_all_profiles(self):
        """Returns all the available profiles."""
        profiles = []
        names = self._get_profile_names()
        for name in names:
            profiles.append(self.get_profile(name))
        return profiles

    def create_profile(self, name, server, username, ignore_ssl_errors):
        """Creates a new profile if one does not already exist for that name."""
        try:
            self.get_profile(name)
        except NoConfigProfileError as ex:
            if name is not None and name != self.DEFAULT_VALUE:
                self._create_profile_section(name)
            else:
                raise ex

        profile = self.get_profile(name)
        self.update_profile(profile.name, server, username, ignore_ssl_errors)
        self._try_complete_setup(profile)

    def update_profile(self, name, server=None, username=None, ignore_ssl_errors=None):
        profile = self.get_profile(name)
        if server:
            self._set_authority_url(server, profile)
        if username:
            self._set_username(username, profile)
        if ignore_ssl_errors is not None:
            self._set_ignore_ssl_errors(ignore_ssl_errors, profile)
        self._save()

    def switch_default_profile(self, new_default_name):
        """Changes what is marked as the default profile in the internal section."""
        if self.get_profile(new_default_name) is None:
            raise NoConfigProfileError(new_default_name)
        self._internal[self.DEFAULT_PROFILE] = new_default_name
        self._save()
        get_main_cli_logger().print_info(
            u"{} has been set as the default profile.".format(new_default_name)
        )

    def delete_profile(self, name):
        """Deletes a profile."""
        if self.get_profile(name) is None:
            raise NoConfigProfileError(name)
        self.parser.remove_section(name)
        if name == self._default_profile_name:
            self._internal[self.DEFAULT_PROFILE] = self.DEFAULT_VALUE
        self._save()

    def _set_authority_url(self, new_value, profile):
        profile[self.AUTHORITY_KEY] = new_value.strip()

    def _set_username(self, new_value, profile):
        profile[self.USERNAME_KEY] = new_value.strip()

    def _set_ignore_ssl_errors(self, new_value, profile):
        profile[self.IGNORE_SSL_ERRORS_KEY] = str(new_value)

    def _get_sections(self):
        return self.parser.sections()

    def _get_profile(self, name):
        return self.parser[name]

    @property
    def _internal(self):
        return self.parser[self._INTERNAL_SECTION]

    @property
    def _default_profile_name(self):
        return self._internal[self.DEFAULT_PROFILE]

    def _get_profile_names(self):
        names = list(self._get_sections())
        names.remove(self._INTERNAL_SECTION)
        return names

    def _create_internal_section(self):
        self.parser.add_section(self._INTERNAL_SECTION)
        self.parser[self._INTERNAL_SECTION] = {}
        self.parser[self._INTERNAL_SECTION][self.DEFAULT_PROFILE] = self.DEFAULT_VALUE

    def _create_profile_section(self, name):
        self.parser.add_section(name)
        self.parser[name] = {}
        self.parser[name][self.AUTHORITY_KEY] = self.DEFAULT_VALUE
        self.parser[name][self.USERNAME_KEY] = self.DEFAULT_VALUE
        self.parser[name][self.IGNORE_SSL_ERRORS_KEY] = str(False)

    def _save(self):
        util.open_file(self.path, u"w+", lambda file: self.parser.write(file))

    def _try_complete_setup(self, profile):
        authority = profile.get(self.AUTHORITY_KEY)
        username = profile.get(self.USERNAME_KEY)

        authority_valid = authority and authority != self.DEFAULT_VALUE
        username_valid = username and username != self.DEFAULT_VALUE

        if not authority_valid or not username_valid:
            return

        self._save()
        get_main_cli_logger().print_info(u"Successfully saved profile '{}'.".format(profile.name))

        default_profile = self._internal.get(self.DEFAULT_PROFILE)
        if default_profile is None or default_profile == self.DEFAULT_VALUE:
            self.switch_default_profile(profile.name)


config_accessor = ConfigAccessor(ConfigParser())



from datetime import datetime, timedelta
import re

from c42eventextractor.common import convert_datetime_to_timestamp

from code42cli.errors import DateArgumentError

TIMESTAMP_REGEX = re.compile(u"(\d{4}-\d{2}-\d{2})\s*(.*)?")
MAGIC_TIME_REGEX = re.compile(u"(\d+)([dhm])$")


def verify_timestamp_order(min_timestamp, max_timestamp):
    if min_timestamp is None or max_timestamp is None:
        return
    if min_timestamp >= max_timestamp:
        raise DateArgumentError(u"Begin date cannot be after end date")


def parse_min_timestamp(begin_date_str, max_days_back=90):
    dt = _parse_timestamp(begin_date_str, _round_datetime_to_day_start)

    boundary_date = _round_datetime_to_day_start(datetime.utcnow() - timedelta(days=max_days_back))
    if dt < boundary_date:
        raise DateArgumentError(u"'Begin date' must be within {0} days.".format(max_days_back))

    return convert_datetime_to_timestamp(dt)


def parse_max_timestamp(end_date_str):
    dt = _parse_timestamp(end_date_str, _round_datetime_to_day_end)
    return convert_datetime_to_timestamp(dt)


def _parse_timestamp(date_str, rounding_func):
    timestamp_match = TIMESTAMP_REGEX.match(date_str)
    magic_match = MAGIC_TIME_REGEX.match(date_str)

    if timestamp_match:
        date, time = timestamp_match.groups()
        dt = _get_dt_from_date_time_pair(date, time)
        if not time:
            dt = rounding_func(dt)

    elif magic_match:
        num, period = magic_match.groups()
        dt = _get_dt_from_magic_time_pair(num, period)
        if period == u"d":
            dt = rounding_func(dt)

    else:
        raise DateArgumentError()
    return dt


def _get_dt_from_date_time_pair(date, time):
    date_format = u"%Y-%m-%d %H:%M:%S"
    if time:
        time = u"{}:{}:{}".format(*time.split(":") + [u"00", u"00"])
    else:
        time = u"00:00:00"
    date_string = u"{} {}".format(date, time)
    try:
        dt = datetime.strptime(date_string, date_format)
    except ValueError:
        raise DateArgumentError()
    else:
        return dt


def _get_dt_from_magic_time_pair(num, period):
    num = int(num)
    if period == u"d":
        dt = datetime.utcnow() - timedelta(days=num)
    elif period == u"h":
        dt = datetime.utcnow() - timedelta(hours=num)
    elif period == u"m":
        dt = datetime.utcnow() - timedelta(minutes=num)
    else:
        raise DateArgumentError(u"Couldn't parse magic time string: {}{}".format(num, period))
    return dt


def _round_datetime_to_day_start(dt):
    return dt.replace(hour=0, minute=0, second=0, microsecond=0)


def _round_datetime_to_day_end(dt):
    return dt.replace(hour=23, minute=59, second=59, microsecond=999000)



import json

from c42eventextractor import ExtractionHandlers
from py42.sdk.queries.query_filter import QueryFilterTimestampField

import code42cli.errors as errors
from code42cli.date_helper import parse_min_timestamp, parse_max_timestamp, verify_timestamp_order
from code42cli.logger import get_main_cli_logger
from code42cli.cmds.alerts.util import get_alert_details
from code42cli.util import warn_interrupt

logger = get_main_cli_logger()


def begin_date_is_required(args, cursor_store):
    if not args.incremental:
        return True
    is_required = cursor_store and cursor_store.get_stored_cursor_timestamp() is None

    # Ignore begin date when in incremental mode, it is not required, and it was passed an argument.
    if not is_required and args.begin:
        logger.print_and_log_info(
            u"Ignoring --begin value as --incremental was passed and cursor checkpoint exists.\n"
        )
        args.begin = None
    return is_required


def verify_begin_date_requirements(args, cursor_store):
    if begin_date_is_required(args, cursor_store) and not args.begin:
        logger.print_and_log_error(u"'begin date' is required.\n")
        logger.print_bold(u"Try using  '-b' or '--begin'. Use `-h` for more info.\n")
        exit(1)


def create_handlers(output_logger, cursor_store, event_key, sdk=None):
    handlers = ExtractionHandlers()
    handlers.TOTAL_EVENTS = 0

    def handle_error(exception):
        errors.ERRORED = True
        if hasattr(exception, u"response") and hasattr(exception.response, u"text"):
            message = u"{0}: {1}".format(exception, exception.response.text)
        else:
            message = exception
        logger.print_and_log_error(message)

    handlers.handle_error = handle_error

    if cursor_store:
        handlers.record_cursor_position = cursor_store.replace_stored_cursor_timestamp
        handlers.get_cursor_position = cursor_store.get_stored_cursor_timestamp

    @warn_interrupt(
        warning=u"Cancelling operation cleanly to keep checkpoint data accurate. One moment..."
    )
    def handle_response(response):
        response_dict = json.loads(response.text)
        events = response_dict.get(event_key)
        if event_key == u"alerts":
            try:
                events = get_alert_details(sdk, events)
            except Exception as ex:
                handlers.handle_error(ex)
        handlers.TOTAL_EVENTS += len(events)
        for event in events:
            output_logger.info(event)

    handlers.handle_response = handle_response
    return handlers


def exit_if_advanced_query_used_with_other_search_args(args):
    args_dict_copy = args.__dict__.copy()
    for arg in (u"advanced_query", u"format", u"sdk", u"profile"):
        args_dict_copy.pop(arg)
    if any(args_dict_copy.values()):
        logger.print_and_log_error(u"You cannot use --advanced-query with additional search args.")
        exit(1)


def create_time_range_filter(filter_cls, begin_date=None, end_date=None):
    """Creates a filter using the given filter class (must be a subclass of 
        :class:`py42.sdk.queries.query_filter.QueryFilterTimestampField`) and date args. Returns 
        `None` if both begin_date and end_date args are `None`.
        
        Args:
            begin_date: The begin date for the range.
            end_date: The end date for the range.
    """
    if not issubclass(filter_cls, QueryFilterTimestampField):
        raise Exception(u"filter_cls must be a subclass of QueryFilterTimestampField")

    if begin_date and end_date:
        min_timestamp = parse_min_timestamp(begin_date)
        max_timestamp = parse_max_timestamp(end_date)
        verify_timestamp_order(min_timestamp, max_timestamp)
        return filter_cls.in_range(min_timestamp, max_timestamp)

    elif begin_date and not end_date:
        min_timestamp = parse_min_timestamp(begin_date)
        return filter_cls.on_or_after(min_timestamp)

    elif end_date and not begin_date:
        max_timestamp = parse_max_timestamp(end_date)
        return filter_cls.on_or_before(max_timestamp)



from code42cli.cmds.search_shared.enums import SearchArguments, OutputFormat, AlertOutputFormat
from code42cli.args import ArgConfig


def create_search_args(search_for, filter_args):
    search_args = {
        SearchArguments.ADVANCED_QUERY: ArgConfig(
            u"--{}".format(SearchArguments.ADVANCED_QUERY.replace(u"_", u"-")),
            metavar=u"QUERY_JSON",
            help=u"A raw JSON {0} query. "
            u"Useful for when the provided query parameters do not satisfy your requirements.\n"
            u"WARNING: Using advanced queries is incompatible with other query-building args.".format(
                search_for
            ),
        ),
        SearchArguments.BEGIN_DATE: ArgConfig(
            u"-b",
            u"--{}".format(SearchArguments.BEGIN_DATE),
            metavar=u"DATE",
            help=u"The beginning of the date range in which to look for {0}, "
            u"can be a date/time in yyyy-MM-dd (UTC) or yyyy-MM-dd HH:MM:SS (UTC+24-hr time) format "
            u"where the 'time' portion of the string can be partial (e.g. '2020-01-01 12' or '2020-01-01 01:15') "
            u"or a short value representing days (30d), hours (24h) or minutes (15m) from current "
            u"time.".format(search_for),
        ),
        SearchArguments.END_DATE: ArgConfig(
            u"-e",
            u"--{}".format(SearchArguments.END_DATE),
            metavar=u"DATE",
            help=u"The end of the date range in which to look for {0}, "
            u"argument format options are the same as --begin.".format(search_for),
        ),
    }
    format_enum = AlertOutputFormat() if search_for == "alerts" else OutputFormat()
    format_and_incremental_args = {
        u"format": ArgConfig(
            u"-f",
            u"--format",
            choices=format_enum,
            default=format_enum.JSON,
            help=u"The format used for outputting {0}.".format(search_for),
        ),
        u"incremental": ArgConfig(
            u"-i",
            u"--incremental",
            action=u"store_true",
            help=u"Only get {0} that were not previously retrieved.".format(search_for),
        ),
    }
    search_args.update(filter_args)
    search_args.update(format_and_incremental_args)
    return search_args



from py42.exceptions import Py42BadRequestError

from code42cli.cmds.detectionlists.commands import DetectionListSubcommandLoader
from code42cli.bulk import generate_template, run_bulk_process
from code42cli.file_readers import create_csv_reader, create_flat_file_reader
from code42cli.errors import UserAlreadyAddedError, UnknownRiskTagError
from code42cli.cmds.detectionlists.enums import DetectionLists, DetectionListUserKeys, RiskTags
from code42cli.cmds.detectionlists.bulk import BulkDetectionList, BulkHighRiskEmployee
from code42cli.util import get_user_id


def try_handle_user_already_added_error(bad_request_err, username_tried_adding, list_name):
    if _error_is_user_already_added(bad_request_err.response.text):
        raise UserAlreadyAddedError(username_tried_adding, list_name)
    return False


def _error_is_user_already_added(bad_request_error_text):
    return u"User already on list" in bad_request_error_text


class DetectionListHandlers(object):
    """Handlers DTO for passing in specific detection list functions.
    
    Args:
        add (callable): A function that adds an employee to the list.
        remove (callable): A function that removes an employee from the list.
        load_add (callable): A function that loads the add-related `ArgConfig`s.
    """

    def __init__(self, add=None, remove=None, load_add=None):
        self.add_employee = add
        self.remove_employee = remove
        self.load_add_description = load_add

    def add_handler(self, attr_name, handler):
        self.__setattr__(attr_name, handler)


class DetectionList(object):
    """An object representing a Code42 detection list. Use this class by passing in handlers for 
    adding and removing employees. This class will handle the bulk-related commands and some 
    search_shared help texts.
    
    Args:
        list_name (str or unicode): An option from the DetectionLists enum. For convenience, use one of the 
            given `classmethods`.
        handlers (DetectionListHandlers): A DTO containing implementations for adding / removing 
            users from specific lists.
        cmd_factory (DetectionListSubcommandLoader): A factory that creates detection list commands.
    """

    def __init__(self, list_name, handlers, subcommand_loader=None):
        self.name = list_name
        self.handlers = handlers
        self.subcommand_loader = subcommand_loader or DetectionListSubcommandLoader(list_name)
        self.bulk_subcommand_loader = self.subcommand_loader.bulk_subcommand_loader
        self.bulk_subcommand_loader.load_commands = lambda: self._load_bulk_subcommands

    @classmethod
    def create_high_risk_employee_list(cls, handlers):
        """Creates a high risk employee detection list.
        
        Args:
            handlers (DetectionListHandlers): A DTO containing implementations for adding / 
                removing users from specific lists.
        
        Returns:
            DetectionList: A high risk employee detection list.
        """
        return cls(DetectionLists.HIGH_RISK_EMPLOYEE, handlers)

    @classmethod
    def create_departing_employee_list(cls, handlers):
        """Creates a departing employee detection list.
        
        Args:
            handlers (DetectionListHandlers): A DTO containing implementations for adding /
                removing users from specific lists.
        
        Returns:
            DetectionList: A departing employee detection list.
        """
        return cls(DetectionLists.DEPARTING_EMPLOYEE, handlers)

    def load_subcommands(self):
        """Loads high risk employee related subcommands"""
        bulk = self.subcommand_loader.create_bulk_command()
        bulk.subcommand_loader.load_commands = lambda: self._load_bulk_subcommands()
        add = self.subcommand_loader.create_add_command(
            self.handlers.add_employee, self.handlers.load_add_description
        )
        remove = self.subcommand_loader.create_remove_command(
            self.handlers.remove_employee, load_username_description
        )
        return [bulk, add, remove]

    def _load_bulk_subcommands(self):
        add = self.bulk_subcommand_loader.create_bulk_add_command(
            self.bulk_add_employees, self.handlers.add_employee
        )
        remove = self.bulk_subcommand_loader.create_bulk_remove_command(self.bulk_remove_employees)
        commands = [add, remove]

        if self.name == DetectionLists.HIGH_RISK_EMPLOYEE:
            commands.extend(self._get_risk_tags_bulk_subcommands())
        else:
            generate_template_cmd = self.bulk_subcommand_loader.create_bulk_generate_template_command(
                self.generate_template_file
            )
            commands.append(generate_template_cmd)
        return commands

    def _get_risk_tags_bulk_subcommands(self):
        bulk_add_risk_tags = self.bulk_subcommand_loader.create_bulk_add_risk_tags_command(
            self.bulk_add_risk_tags, add_risk_tags
        )
        bulk_remove_risk_tags = self.bulk_subcommand_loader.create_bulk_remove_risk_tags_command(
            self.bulk_remove_risk_tags, remove_risk_tags
        )

        self.handlers.add_handler(u"add_risk_tags", add_risk_tags)
        self.handlers.add_handler(u"remove_risk_tags", remove_risk_tags)
        generate_template_cmd = self.bulk_subcommand_loader.create_hre_bulk_generate_template_command(
            self.generate_template_file
        )
        return [bulk_add_risk_tags, bulk_remove_risk_tags, generate_template_cmd]

    def generate_template_file(self, cmd, path=None):
        """Generates a template file a user would need to fill-in for bulk operating on the 
        detection list.

        Args:
            cmd (str or unicode): An option from the `BulkCommandType` enum specifying which type of file to 
                generate.
            path (str or unicode, optional): A path to put the file after it's generated. If None, will use 
                the current working directory. Defaults to None.
        """

        if self.name == DetectionLists.HIGH_RISK_EMPLOYEE:
            detection_list = BulkHighRiskEmployee()
        else:
            detection_list = BulkDetectionList()
        handler = detection_list.get_handler(self.handlers, cmd)
        generate_template(handler, path)

    def bulk_add_employees(self, sdk, profile, csv_file):
        """Takes a csv file with each row representing an employee and adds them all to a 
        detection list in a bulk fashion.

        Args:
            sdk (py42.sdk.SDKClient): The py42 sdk.
            profile (Code42Profile): The profile under which to execute this command.
            csv_file (str or unicode): The path to the csv file containing rows of users.
        """
        reader = create_csv_reader(csv_file)
        run_bulk_process(lambda **kwargs: self._add_employee(sdk, profile, **kwargs), reader)

    def bulk_remove_employees(self, sdk, profile, users_file):
        """Takes a flat file with each row containing a username and removes them all from the 
        detection list in a bulk fashion.
        
        Args:
            sdk (py42.sdk.SDKClient): The py42 sdk.
            profile (Code42Profile): The profile under which to execute this command.
            users_file (str or unicode): The path to the file containing rows of user names.
        """
        reader = create_flat_file_reader(users_file)
        run_bulk_process(
            lambda *args, **kwargs: self._remove_employee(sdk, profile, *args, **kwargs), reader
        )

    def _add_employee(self, sdk, profile, **kwargs):
        self.handlers.add_employee(sdk, profile, **kwargs)

    def _remove_employee(self, sdk, profile, *args, **kwargs):
        self.handlers.remove_employee(sdk, profile, *args, **kwargs)

    def bulk_add_risk_tags(self, sdk, profile, csv_file):
        reader = create_csv_reader(csv_file)
        run_bulk_process(lambda **kwargs: add_risk_tags(sdk, profile, **kwargs), reader)

    def bulk_remove_risk_tags(self, sdk, profile, csv_file):
        reader = create_csv_reader(csv_file)
        run_bulk_process(lambda **kwargs: remove_risk_tags(sdk, profile, **kwargs), reader)


def load_username_description(argument_collection):
    """Loads the arg descriptions for the `username` CLI parameter."""
    username = argument_collection.arg_configs[DetectionListUserKeys.USERNAME]
    username.set_help(u"A Code42 username for an employee.")


def load_user_descriptions(argument_collection):
    """Loads the arg descriptions related to updating fields about a detection list user, such as 
    notes or a cloud alias.
    
    Args:
        argument_collection (ArgConfigCollection): The arg configs off the command that needs its 
            user descriptions loaded.
    """
    load_username_description(argument_collection)
    cloud_alias = argument_collection.arg_configs[DetectionListUserKeys.CLOUD_ALIAS]
    notes = argument_collection.arg_configs[DetectionListUserKeys.NOTES]
    cloud_alias.set_help(u"An alternative email address for another cloud service.")
    notes.set_help(u"Notes about the employee.")


def update_user(sdk, user_id, cloud_alias=None, risk_tag=None, notes=None):
    """Updates a detection list user.
    
    Args:
        sdk (py42.sdk.SDKClient): py42 sdk.
        user_id (str or unicode): The ID of the user to update. This is their `userUid` found from 
            `sdk.users.get_by_username()`.
        cloud_alias (str or unicode): A cloud alias to add to the user.
        risk_tag (iter[str or unicode]): A list of risk tags associated with user.
        notes (str or unicode): Notes about the user.
    """
    if cloud_alias:
        sdk.detectionlists.add_user_cloud_alias(user_id, cloud_alias)
    if risk_tag:
        try_add_risk_tags(sdk, user_id, risk_tag)
    if notes:
        sdk.detectionlists.update_user_notes(user_id, notes)


def try_add_risk_tags(sdk, user_id, risk_tag):
    _try_add_or_remove_risk_tags(user_id, risk_tag, sdk.detectionlists.add_user_risk_tags)


def try_remove_risk_tags(sdk, user_id, risk_tag):
    _try_add_or_remove_risk_tags(user_id, risk_tag, sdk.detectionlists.remove_user_risk_tags)


def _try_add_or_remove_risk_tags(user_id, risk_tag, func):
    try:
        func(user_id, risk_tag)
    except Py42BadRequestError:
        _try_handle_bad_risk_tag(risk_tag)
        raise


def _try_handle_bad_risk_tag(tags):
    options = list(RiskTags())
    unknowns = [tag for tag in tags if tag not in options] if tags else None
    if unknowns:
        raise UnknownRiskTagError(unknowns)


def handle_list_args(list_arg):
    """Converts str args to a list. Useful for `bulk` commands which don't use `argparse` but
    instead pass in values from files, such as in the form "item1 item2"."""
    if list_arg and not isinstance(list_arg, list):
        return list_arg.split()
    return list_arg


def add_risk_tags(sdk, profile, username, tag):
    risk_tag = handle_list_args(tag)
    user_id = get_user_id(sdk, username)
    try_add_risk_tags(sdk, user_id, risk_tag)


def remove_risk_tags(sdk, profile, username, tag):
    risk_tag = handle_list_args(tag)
    user_id = get_user_id(sdk, username)
    try_remove_risk_tags(sdk, user_id, risk_tag)



from threading import Thread, Lock

from py42.exceptions import Py42HTTPError, Py42ForbiddenError

from code42cli.errors import Code42CLIError
from code42cli.compat import queue
from code42cli.logger import get_main_cli_logger


class WorkerStats(object):
    """Stats about the tasks that have run."""

    def __init__(self, total):
        self.total = total

    _total_processed = 0
    _total_errors = 0
    __total_processed_lock = Lock()
    __total_errors_lock = Lock()

    @property
    def total_processed(self):
        """The total number of tasks executed."""
        return self._total_processed

    @property
    def total_errors(self):
        """The amount of errors that occurred."""
        return self._total_errors

    @property
    def total_successes(self):
        val = self._total_processed - self._total_errors
        return val if val >= 0 else 0

    def __str__(self):
        return u"{0} succeeded, {1} failed out of {2}".format(
            self.total_successes, self._total_errors, self.total
        )

    def increment_total_processed(self):
        """+1 to self.total_processed"""
        with self.__total_processed_lock:
            self._total_processed += 1

    def increment_total_errors(self):
        """+1 to self.total_errors"""
        with self.__total_errors_lock:
            self._total_errors += 1


class Worker(object):
    def __init__(self, thread_count, expected_total):
        self._queue = queue.Queue()
        self._thread_count = thread_count
        self._stats = WorkerStats(expected_total)
        self.__started = False
        self.__start_lock = Lock()

    def do_async(self, func, *args, **kwargs):
        """Execute the given func asynchronously given *args and **kwargs.
        
        Args:
            func (callable): The function to execute asynchronously.
            *args (iter): Positional args to pass to the function.
            **kwargs (dict): Key-value args to pass to the function.
        """
        if not self.__started:
            with self.__start_lock:
                if not self.__started:
                    self.__start()
                    self.__started = True
        self._queue.put({u"func": func, u"args": args, u"kwargs": kwargs})

    @property
    def stats(self):
        """Stats about the tasks that have been executed, such as the total errors that occurred.
        """
        return self._stats

    def wait(self):
        """Wait for the tasks in the queue to complete. This should usually be called before 
        program termination."""
        self._queue.join()

    def _process_queue(self):
        while True:
            try:
                task = self._queue.get()
                func = task[u"func"]
                args = task[u"args"]
                kwargs = task[u"kwargs"]
                func(*args, **kwargs)
            except Code42CLIError as err:
                self._increment_total_errors()
                logger = get_main_cli_logger()
                logger.log_error(err)
            except Py42ForbiddenError as err:
                self._increment_total_errors()
                logger = get_main_cli_logger()
                logger.log_verbose_error(http_request=err.response.request)
                logger.log_permissions_error()
            except Py42HTTPError as err:
                self._increment_total_errors()
                logger = get_main_cli_logger()
                logger.log_verbose_error(http_request=err.response.request)
            except Exception:
                self._increment_total_errors()
                logger = get_main_cli_logger()
                logger.log_verbose_error()
            finally:
                self._stats.increment_total_processed()
                self._queue.task_done()

    def __start(self):
        for _ in range(0, self._thread_count):
            t = Thread(target=self._process_queue)
            t.daemon = True
            t.start()

    def _increment_total_errors(self):
        self._stats.increment_total_errors()



from getpass import getpass

from code42cli import MAIN_COMMAND
import code42cli.profile as cliprofile
from code42cli.compat import str
from code42cli.profile import print_and_log_no_existing_profile
from code42cli.args import PROFILE_HELP
from code42cli.commands import Command, SubcommandLoader
from code42cli.sdk_client import validate_connection
from code42cli.util import does_user_agree
from code42cli.logger import get_main_cli_logger


class ProfileSubcommandLoader(SubcommandLoader):
    SHOW = u"show"
    LIST = u"list"
    USE = u"use"
    RESET_PW = u"reset-pw"
    CREATE = u"create"
    UPDATE = u"update"
    DELETE = u"delete"
    DELETE_ALL = u"delete-all"

    def load_commands(self):
        """Sets up the `profile` subcommand with all of its subcommands."""
        usage_prefix = u"{} profile".format(MAIN_COMMAND)

        show = Command(
            self.SHOW,
            u"Print the details of a profile.",
            u"{} {}".format(usage_prefix, u"show <optional-args>"),
            handler=show_profile,
            arg_customizer=_load_optional_profile_description,
        )

        list_all = Command(
            self.LIST,
            u"Show all existing stored profiles.",
            u"{} {}".format(usage_prefix, u"list"),
            handler=list_profiles,
        )

        use = Command(
            self.USE,
            u"Set a profile as the default.",
            u"{} {}".format(usage_prefix, u"use <profile-name>"),
            handler=use_profile,
        )

        reset_pw = Command(
            self.RESET_PW,
            u"Change the stored password for a profile.",
            u"{} {}".format(usage_prefix, u"reset-pw <optional-args>"),
            handler=prompt_for_password_reset,
            arg_customizer=_load_optional_profile_description,
        )

        create = Command(
            self.CREATE,
            u"Create profile settings. The first profile created will be the default.",
            u"{} {}".format(
                usage_prefix,
                u"create --name <profile-name> --server <server-address> --username <username>",
            ),
            handler=create_profile,
            arg_customizer=_load_profile_create_descriptions,
        )

        update = Command(
            self.UPDATE,
            u"Update an existing profile.",
            u"{} {}".format(usage_prefix, u"update <optional args>"),
            handler=update_profile,
            arg_customizer=_load_profile_update_descriptions,
        )

        delete = Command(
            self.DELETE,
            u"Deletes a profile and its stored password (if any).",
            u"{} {}".format(usage_prefix, u"delete <profile-name>"),
            handler=delete_profile,
        )

        delete_all = Command(
            self.DELETE_ALL,
            u"Deletes all profiles and saved passwords (if any).",
            u"{} {}".format(usage_prefix, u"delete-all"),
            handler=delete_all_profiles,
        )

        return [show, list_all, use, reset_pw, create, update, delete, delete_all]


def show_profile(name=None):
    """Prints the given profile to stdout."""
    c42profile = cliprofile.get_profile(name)
    logger = get_main_cli_logger()
    logger.print_info(u"\n{0}:".format(c42profile.name))
    logger.print_info(u"\t* username = {}".format(c42profile.username))
    logger.print_info(u"\t* authority url = {}".format(c42profile.authority_url))
    logger.print_info(u"\t* ignore-ssl-errors = {}".format(c42profile.ignore_ssl_errors))
    if cliprofile.get_stored_password(c42profile.name) is not None:
        logger.print_info(u"\t* A password is set.")
    logger.print_info(u"")


def create_profile(name, server, username, disable_ssl_errors=False):
    cliprofile.create_profile(name, server, username, disable_ssl_errors)
    _prompt_for_allow_password_set(name)
    get_main_cli_logger().print_info(u"Successfully created profile '{}'.".format(name))


def update_profile(name=None, server=None, username=None, disable_ssl_errors=None):
    profile = cliprofile.get_profile(name)
    cliprofile.update_profile(profile.name, server, username, disable_ssl_errors)
    _prompt_for_allow_password_set(profile.name)
    get_main_cli_logger().print_info(u"Profile '{}' has been updated.".format(profile.name))


def prompt_for_password_reset(name=None):
    """Securely prompts for your password and then stores it using keyring."""
    c42profile = cliprofile.get_profile(name)
    new_password = getpass()
    _validate_connection(c42profile.authority_url, c42profile.username, new_password)
    cliprofile.set_password(new_password, c42profile.name)


def _validate_connection(authority, username, password):
    if not validate_connection(authority, username, password):
        logger = get_main_cli_logger()
        logger.print_and_log_error(
            u"Your credentials failed to validate, so your password was not stored."
            u"Check your network connection and the spelling of your username and server URL."
        )
        exit(1)


def list_profiles(*args):
    """Lists all profiles that exist for this OS user."""
    profiles = cliprofile.get_all_profiles()
    logger = get_main_cli_logger()
    if not profiles:
        print_and_log_no_existing_profile()
        return
    for profile in profiles:
        logger.print_info(str(profile))


def use_profile(name):
    """Changes the default profile to the given one."""
    cliprofile.switch_default_profile(name)


def delete_profile(name):
    logger = get_main_cli_logger()
    if cliprofile.is_default_profile(name):
        logger.print_info(u"\n{} is currently the default profile!".format(name))
    if not does_user_agree(
        u"\nDeleting this profile will also delete any stored passwords and checkpoints. "
        u"Are you sure? (y/n): "
    ):
        return
    cliprofile.delete_profile(name)


def delete_all_profiles():
    existing_profiles = cliprofile.get_all_profiles()
    logger = get_main_cli_logger()
    if existing_profiles:
        logger.print_info(u"\nAre you sure you want to delete the following profiles?")
        for profile in existing_profiles:
            logger.print_info(u"\t{}".format(profile.name))
        if does_user_agree(
            u"\nThis will also delete any stored passwords and checkpoints. (y/n): "
        ):
            for profile in existing_profiles:
                cliprofile.delete_profile(profile.name)
    else:
        logger.print_info(u"\nNo profiles exist. Nothing to delete.")


def _load_optional_profile_description(argument_collection):
    profile = argument_collection.arg_configs[u"name"]
    profile.add_short_option_name(u"-n")
    profile.set_help(PROFILE_HELP)


def _load_profile_create_descriptions(argument_collection):
    profile = argument_collection.arg_configs[u"name"]
    profile.set_help(PROFILE_HELP)
    profile.add_short_option_name(u"-n")
    argument_collection.arg_configs[u"server"].add_short_option_name(u"-s")
    argument_collection.arg_configs[u"username"].add_short_option_name(u"-u")
    _load_profile_settings_descriptions(argument_collection)


def _load_profile_update_descriptions(argument_collection):
    _load_optional_profile_description(argument_collection)
    _load_profile_settings_descriptions(argument_collection)
    argument_collection.arg_configs[u"server"].add_short_option_name(u"-s")
    argument_collection.arg_configs[u"username"].add_short_option_name(u"-u")


def _load_profile_settings_descriptions(argument_collection):
    server = argument_collection.arg_configs[u"server"]
    username = argument_collection.arg_configs[u"username"]
    disable_ssl_errors = argument_collection.arg_configs[u"disable_ssl_errors"]
    server.set_help(u"The url and port of the Code42 server.")
    username.set_help(u"The username of the Code42 API user.")
    disable_ssl_errors.set_help(
        u"For development purposes, do not validate the SSL certificates of Code42 servers. "
        u"This is not recommended unless it is required."
    )


def _prompt_for_allow_password_set(profile_name):
    if does_user_agree(u"Would you like to set a password? (y/n): "):
        prompt_for_password_reset(profile_name)



import inspect

from code42cli.bulk import BulkCommandType
from code42cli.commands import Command, SubcommandLoader
from code42cli.cmds.detectionlists.bulk import HighRiskBulkCommandType


def create_usage_prefix(detection_list_name):
    return u"code42 {}".format(detection_list_name)


def create_bulk_usage_prefix(detection_list_name):
    return u"{} bulk".format(create_usage_prefix(detection_list_name))


def _load_bulk_generate_template_description(argument_collection):
    cmd_type = argument_collection.arg_configs[u"cmd"]
    cmd_type.set_help(u"The type of command the template with be used for.")
    cmd_type.set_choices(BulkCommandType())


class DetectionListSubcommandLoader(SubcommandLoader):
    BULK = u"bulk"
    ADD = BulkCommandType.ADD
    REMOVE = BulkCommandType.REMOVE
    _USAGE_SUFFIX = u"<username> <optional args>"

    def __init__(self, detection_list_name, bulk_subcommand_loader=None):
        super(DetectionListSubcommandLoader, self).__init__(detection_list_name)
        self._name = detection_list_name
        self._usage_prefix = create_usage_prefix(detection_list_name)
        self.bulk_subcommand_loader = bulk_subcommand_loader or DetectionListBulkSubcommandLoader(
            self.BULK, detection_list_name
        )

    def create_bulk_command(self):
        return Command(
            self.BULK,
            u"Tools for executing bulk {} commands.".format(self._name),
            subcommand_loader=self.bulk_subcommand_loader,
        )

    def create_add_command(self, handler, arg_customizer):
        return Command(
            self.ADD,
            u"Add a user to the {} detection list.".format(self._name),
            u"{} {} {}".format(self._usage_prefix, BulkCommandType.ADD, self._USAGE_SUFFIX),
            handler=handler,
            arg_customizer=arg_customizer,
        )

    def create_remove_command(self, handler, arg_customizer):
        return Command(
            self.REMOVE,
            u"Remove a user from the {} detection list.".format(self._name),
            u"{} {} {}".format(self._usage_prefix, BulkCommandType.REMOVE, self._USAGE_SUFFIX),
            handler=handler,
            arg_customizer=arg_customizer,
        )


class DetectionListBulkSubcommandLoader(SubcommandLoader):
    ADD = BulkCommandType.ADD
    REMOVE = BulkCommandType.REMOVE
    GENERATE_TEMPLATE = u"generate-template"

    def __init__(self, root_command_name, detection_list_name):
        super(DetectionListBulkSubcommandLoader, self).__init__(root_command_name)
        self._bulk_usage_prefix = create_bulk_usage_prefix(detection_list_name)
        self._name = detection_list_name

    def create_bulk_generate_template_command(self, handler):
        return Command(
            self.GENERATE_TEMPLATE,
            u"Generate the necessary csv template needed for bulk adding users.",
            u"{} generate-template <cmd> <optional args>".format(self._bulk_usage_prefix),
            handler=handler,
            arg_customizer=_load_bulk_generate_template_description,
        )

    def create_hre_bulk_generate_template_command(self, handler):
        return Command(
            u"generate-template",
            u"Generate the necessary csv template for bulk actions.",
            u"{} generate-template <cmd> <optional args>".format(self._bulk_usage_prefix),
            handler=handler,
            arg_customizer=self._load_hre_bulk_generate_template_description,
        )

    def create_bulk_add_command(self, cmd_handler, row_handler):
        file_format = _get_file_format(row_handler)
        return Command(
            self.ADD,
            u"Add users to the {} detection list. CSV file format: `{}`.".format(
                self._name, file_format
            ),
            u"{} {} <filename>".format(self._bulk_usage_prefix, BulkCommandType.ADD),
            handler=cmd_handler,
            arg_customizer=self._load_bulk_add_description,
        )

    def create_bulk_remove_command(self, cmd_handler):
        return Command(
            self.REMOVE,
            u"Remove users from the {} detection list. "
            u"The file format is an end-line-delimited list of users.".format(self._name),
            u"{} {} <file>".format(self._bulk_usage_prefix, BulkCommandType.REMOVE),
            handler=cmd_handler,
            arg_customizer=self._load_bulk_remove_description,
        )

    def create_bulk_add_risk_tags_command(self, cmd_handler, row_handler):
        file_format = _get_file_format(row_handler)
        return Command(
            u"add-risk-tags",
            u"Associates risk tags with a user in bulk. CSV file format: `{}`.".format(file_format),
            u"{} {} <file>".format(self._bulk_usage_prefix, HighRiskBulkCommandType.ADD_RISK_TAG),
            handler=cmd_handler,
            arg_customizer=self._load_bulk_add_risk_tags_description,
        )

    def create_bulk_remove_risk_tags_command(self, cmd_handler, row_handler):
        file_format = _get_file_format(row_handler)
        return Command(
            u"remove-risk-tags",
            u"Disassociates risk tags from a user in bulk. CSV file format: `{}`.".format(
                file_format
            ),
            u"{} {} <file>".format(
                self._bulk_usage_prefix, HighRiskBulkCommandType.REMOVE_RISK_TAG
            ),
            handler=cmd_handler,
            arg_customizer=self._load_bulk_remove_risk_tags_description,
        )

    @staticmethod
    def _load_bulk_generate_template_description(argument_collection):
        cmd_type = argument_collection.arg_configs[u"cmd"]
        cmd_type.set_help(u"The type of command the template will be used for.")
        cmd_type.set_choices(BulkCommandType())

    @staticmethod
    def _load_hre_bulk_generate_template_description(argument_collection):
        cmd_type = argument_collection.arg_configs[u"cmd"]
        cmd_type.set_help(u"The type of command the template will be used for.")
        cmd_type.set_choices(HighRiskBulkCommandType())

    def _load_bulk_add_description(self, argument_collection):
        csv_file = argument_collection.arg_configs[u"csv_file"]
        csv_file.set_help(
            u"The path to the csv file for bulk adding users to the {} detection list.".format(
                self._name
            )
        )

    def _load_bulk_remove_description(self, argument_collection):
        users_file = argument_collection.arg_configs[u"users_file"]
        users_file.set_help(
            u"A file containing a line-separated list of users to remove form the {} detection list.".format(
                self._name
            )
        )

    def _load_bulk_add_risk_tags_description(self, argument_collection):
        csv_file = argument_collection.arg_configs[u"csv_file"]
        csv_file.set_help(
            u"A file containing a ',' separated username with space-separated tags to add "
            u"to the {} detection list. "
            u"e.g. test@email.com,tag1 tag2 tag3".format(self._name)
        )

    def _load_bulk_remove_risk_tags_description(self, argument_collection):
        csv_file = argument_collection.arg_configs[u"csv_file"]
        csv_file.set_help(
            u"A file containing a ',' separated username with space-separated tags to remove "
            u"from the {} detection list. "
            u"e.g. test@email.com,tag1 tag2 tag3".format(self._name)
        )


def _get_file_format(row_handler):
    args = inspect.getargspec(row_handler).args
    args.remove(u"profile")
    args.remove(u"sdk")
    return u", ".join(args)



from code42cli.args import ArgConfig
from code42cli.cmds.search_shared import logger_factory, args
from code42cli.cmds.search_shared.enums import (
    FileEventFilterArguments,
    ServerProtocol,
    ExposureType,
)
from code42cli.cmds.securitydata.extraction import extract
from code42cli.cmds.search_shared.cursor_store import FileEventCursorStore
from code42cli.commands import Command, SubcommandLoader


class SecurityDataSubcommandLoader(SubcommandLoader):
    PRINT = u"print"
    WRITE_TO = u"write-to"
    SEND_TO = u"send-to"
    CLEAR_CHECKPOINT = u"clear-checkpoint"

    def load_commands(self):
        """Sets up the `security-data` subcommand with all of its subcommands."""
        usage_prefix = u"code42 security-data"

        print_func = Command(
            self.PRINT,
            u"Print file events to stdout.",
            u"{} {}".format(usage_prefix, u"print <optional-args>"),
            handler=print_out,
            arg_customizer=_load_search_args,
            use_single_arg_obj=True,
        )

        write = Command(
            self.WRITE_TO,
            u"Write file events to the file with the given name.",
            u"{} {}".format(usage_prefix, u"write-to <filename> <optional-args>"),
            handler=write_to,
            arg_customizer=_load_write_to_args,
            use_single_arg_obj=True,
        )

        send = Command(
            self.SEND_TO,
            u"Send file events to the given server address.",
            u"{} {}".format(usage_prefix, u"send-to <server-address> <optional-args>"),
            handler=send_to,
            arg_customizer=_load_send_to_args,
            use_single_arg_obj=True,
        )

        clear = Command(
            self.CLEAR_CHECKPOINT,
            u"Remove the saved file event checkpoint from 'incremental' (-i) mode.",
            u"{} {}".format(usage_prefix, u"clear-checkpoint <optional-args>"),
            handler=clear_checkpoint,
        )

        return [print_func, write, send, clear]


def clear_checkpoint(sdk, profile):
    """Removes the stored checkpoint that keeps track of the last file event retrieved for the given profile.
        To use, run `code42 security-data clear-checkpoint`.
        This affects `incremental` mode by causing it to behave like it has never been run before.
    """
    FileEventCursorStore(profile.name).replace_stored_cursor_timestamp(None)


def print_out(sdk, profile, args):
    """Activates 'print' command. It gets security events and prints them to stdout."""
    logger = logger_factory.get_logger_for_stdout(args.format)
    extract(sdk, profile, logger, args)


def write_to(sdk, profile, args):
    """Activates 'write-to' command. It gets security events and writes them to the given file."""
    logger = logger_factory.get_logger_for_file(args.output_file, args.format)
    extract(sdk, profile, logger, args)


def send_to(sdk, profile, args):
    """Activates 'send-to' command. It gets security events and logs them to the given server."""
    logger = logger_factory.get_logger_for_server(args.server, args.protocol, args.format)
    extract(sdk, profile, logger, args)


def _load_write_to_args(arg_collection):
    output_file = ArgConfig(u"output_file", help=u"The name of the local file to send output to.")
    arg_collection.append(u"output_file", output_file)
    _load_search_args(arg_collection)


def _load_send_to_args(arg_collection):
    send_to_args = {
        u"server": ArgConfig(u"server", help=u"The server address to send output to."),
        u"protocol": ArgConfig(
            u"-p",
            u"--protocol",
            choices=ServerProtocol(),
            default=ServerProtocol.UDP,
            help=u"Protocol used to send logs to server.",
        ),
    }

    arg_collection.extend(send_to_args)
    _load_search_args(arg_collection)


def _load_search_args(arg_collection):
    filter_args = {
        FileEventFilterArguments.EXPOSURE_TYPES: ArgConfig(
            u"-t",
            u"--{}".format(FileEventFilterArguments.EXPOSURE_TYPES),
            nargs=u"+",
            help=u"Limits events to those with given exposure types. "
            u"Available choices={0}".format(list(ExposureType())),
        ),
        FileEventFilterArguments.C42_USERNAME: ArgConfig(
            u"--{}".format(FileEventFilterArguments.C42_USERNAME.replace(u"_", u"-")),
            nargs=u"+",
            help=u"Limits events to endpoint events for these users.",
        ),
        FileEventFilterArguments.ACTOR: ArgConfig(
            u"--{}".format(FileEventFilterArguments.ACTOR),
            nargs=u"+",
            help=u"Limits events to only those enacted by the cloud service user of the person who caused the event.",
        ),
        FileEventFilterArguments.MD5: ArgConfig(
            u"--{}".format(FileEventFilterArguments.MD5),
            nargs=u"+",
            help=u"Limits events to file events where the file has one of these MD5 hashes.",
        ),
        FileEventFilterArguments.SHA256: ArgConfig(
            u"--{}".format(FileEventFilterArguments.SHA256),
            nargs=u"+",
            action=u"store",
            help=u"Limits events to file events where the file has one of these SHA256 hashes.",
        ),
        FileEventFilterArguments.SOURCE: ArgConfig(
            u"--{}".format(FileEventFilterArguments.SOURCE),
            nargs=u"+",
            help=u"Limits events to only those from one of these sources. Example=Gmail.",
        ),
        FileEventFilterArguments.FILE_NAME: ArgConfig(
            u"--{}".format(FileEventFilterArguments.FILE_NAME.replace(u"_", u"-")),
            nargs=u"+",
            help=u"Limits events to file events where the file has one of these names.",
        ),
        FileEventFilterArguments.FILE_PATH: ArgConfig(
            u"--{}".format(FileEventFilterArguments.FILE_PATH.replace(u"_", u"-")),
            nargs=u"+",
            help=u"Limits events to file events where the file is located at one of these paths.",
        ),
        FileEventFilterArguments.PROCESS_OWNER: ArgConfig(
            u"--{}".format(FileEventFilterArguments.PROCESS_OWNER.replace(u"_", u"-")),
            nargs=u"+",
            help=u"Limits events to exposure events where one of these users "
            u"owns the process behind the exposure.",
        ),
        FileEventFilterArguments.TAB_URL: ArgConfig(
            u"--{}".format(FileEventFilterArguments.TAB_URL.replace(u"_", u"-")),
            nargs=u"+",
            help=u"Limits events to be exposure events with one of these destination tab URLs.",
        ),
        FileEventFilterArguments.INCLUDE_NON_EXPOSURE_EVENTS: ArgConfig(
            u"--include-non-exposure",
            action=u"store_true",
            help=u"Get all events including non-exposure events.",
        ),
    }
    search_args = args.create_search_args(search_for=u"file events", filter_args=filter_args)
    arg_collection.extend(search_args)



from code42cli import MAIN_COMMAND
from code42cli.main import MainSubcommandLoader


def _get_matches(current, options):
    matches = []
    current = current.strip()
    for opt in options:
        if opt.startswith(current) and opt != current:
            matches.append(opt)
    return matches


def _get_next_full_set_of_commands(cmd_loader, current):
    cmd_loader = cmd_loader.subtrees[current]
    return cmd_loader.names


class Completer(object):
    def __init__(self, main_cmd_loader=None):
        self._main_cmd_loader = main_cmd_loader or MainSubcommandLoader(u"")

    def complete(self, cmdline, point=None):
        try:
            point = point or len(cmdline)
            args = cmdline[0:point].split()
            if len(args) < 2:
                # `code42` already completes w/o
                return self._main_cmd_loader.names if args[0] == MAIN_COMMAND else []

            current = args[-1]
            cmd_loader = self._search_trees(args)
            if not cmd_loader:
                return []

            options = cmd_loader.names
            if current in options:
                # `current` is already complete
                return _get_next_full_set_of_commands(cmd_loader, current)

            return _get_matches(current, options) if options else []
        except:
            return []

    def _search_trees(self, args):
        # Find cmd_loader at lowest level from given args
        cmd_loader = self._main_cmd_loader
        if len(args) > 2:
            for arg in args[1:-1]:
                cmd_loader = cmd_loader.subtrees[arg]
        return cmd_loader


def complete(cmdline, point):
    choices = Completer().complete(cmdline, point)
    print(u" \n".join(choices))



IS_INCREMENTAL_KEY = u"incremental"


class OutputFormat(object):
    CEF = u"CEF"
    JSON = u"JSON"
    RAW = u"RAW-JSON"

    def __iter__(self):
        return iter([self.CEF, self.JSON, self.RAW])


class AlertOutputFormat(object):
    JSON = u"JSON"
    RAW = u"RAW-JSON"

    def __iter__(self):
        return iter([self.JSON, self.RAW])


class AlertSeverity(object):
    HIGH = u"HIGH"
    MEDIUM = u"MEDIUM"
    LOW = u"LOW"

    def __iter__(self):
        return iter(self._as_list())

    def __len__(self):
        return len(self._as_list())

    def _as_list(self):
        return [self.HIGH, self.MEDIUM, self.LOW]


class AlertState(object):
    OPEN = u"OPEN"
    DISMISSED = u"RESOLVED"

    def __iter__(self):
        return iter(self._as_list())

    def __len__(self):
        return len(self._as_list())

    def _as_list(self):
        return [self.OPEN, self.DISMISSED]


class ExposureType(object):
    SHARED_VIA_LINK = u"SharedViaLink"
    SHARED_TO_DOMAIN = u"SharedToDomain"
    APPLICATION_READ = u"ApplicationRead"
    CLOUD_STORAGE = u"CloudStorage"
    REMOVABLE_MEDIA = u"RemovableMedia"
    IS_PUBLIC = u"IsPublic"

    def __iter__(self):
        return iter(self._as_list())

    def __len__(self):
        return len(self._as_list())

    def _as_list(self):
        return [
            self.SHARED_VIA_LINK,
            self.SHARED_TO_DOMAIN,
            self.APPLICATION_READ,
            self.CLOUD_STORAGE,
            self.REMOVABLE_MEDIA,
            self.IS_PUBLIC,
        ]


class RuleType(object):
    ENDPOINT_EXFILTRATION = u"FedEndpointExfiltration"
    CLOUD_SHARE_PERMISSIONS = u"FedCloudSharePermissions"
    FILE_TYPE_MISMATCH = u"FedFileTypeMismatch"

    def __iter__(self):
        return iter(self._as_list())

    def __len__(self):
        return len(self._as_list())

    def _as_list(self):
        return [self.ENDPOINT_EXFILTRATION, self.CLOUD_SHARE_PERMISSIONS, self.FILE_TYPE_MISMATCH]


class ServerProtocol(object):
    TCP = u"TCP"
    UDP = u"UDP"

    def __iter__(self):
        return iter([self.TCP, self.UDP])


class SearchArguments(object):
    """These string values should match `argparse` stored parameter names. For example, for the 
    CLI argument `--c42-username`, the string should be `c42_username`."""

    ADVANCED_QUERY = u"advanced_query"
    BEGIN_DATE = u"begin"
    END_DATE = u"end"

    def __iter__(self):
        return iter([self.ADVANCED_QUERY, self.BEGIN_DATE, self.END_DATE])


class FileEventFilterArguments(SearchArguments):
    EXPOSURE_TYPES = u"type"
    C42_USERNAME = u"c42_username"
    ACTOR = u"actor"
    MD5 = u"md5"
    SHA256 = u"sha256"
    SOURCE = u"source"
    FILE_NAME = u"file_name"
    FILE_PATH = u"file_path"
    PROCESS_OWNER = u"process_owner"
    TAB_URL = u"tab_url"
    INCLUDE_NON_EXPOSURE_EVENTS = u"include_non_exposure"

    def __iter__(self):
        return iter(
            [
                self.EXPOSURE_TYPES,
                self.C42_USERNAME,
                self.ACTOR,
                self.MD5,
                self.SHA256,
                self.SOURCE,
                self.FILE_NAME,
                self.FILE_PATH,
                self.PROCESS_OWNER,
                self.TAB_URL,
                self.INCLUDE_NON_EXPOSURE_EVENTS,
            ]
        )


class AlertFilterArguments(object):
    STATE = u"state"
    SEVERITY = u"severity"
    ACTOR = u"actor"
    ACTOR_CONTAINS = u"actor_contains"
    EXCLUDE_ACTOR = u"exclude_actor"
    EXCLUDE_ACTOR_CONTAINS = u"exclude_actor_contains"
    RULE_NAME = u"rule_name"
    EXCLUDE_RULE_NAME = u"exclude_rule_name"
    RULE_ID = u"rule_id"
    EXCLUDE_RULE_ID = u"exclude_rule_id"
    RULE_TYPE = u"rule_type"
    EXCLUDE_RULE_TYPE = u"exclude_rule_type"
    DESCRIPTION = u"description"

    def __iter__(self):
        return iter(
            [
                self.STATE,
                self.SEVERITY,
                self.ACTOR,
                self.ACTOR_CONTAINS,
                self.EXCLUDE_ACTOR,
                self.EXCLUDE_ACTOR_CONTAINS,
                self.RULE_NAME,
                self.EXCLUDE_RULE_NAME,
                self.RULE_ID,
                self.EXCLUDE_RULE_ID,
                self.RULE_TYPE,
                self.EXCLUDE_RULE_TYPE,
                self.DESCRIPTION,
            ]
        )



import argparse
from argparse import RawDescriptionHelpFormatter, SUPPRESS

from py42.__version__ import __version__ as py42version

from code42cli.__version__ import __version__ as cliversion

BANNER = u""" 
 dP""b8  dP"Yb  8888b. 888888  dP88  oP"Yb. 
dP   `" dP   Yb 8I  Yb 88__   dP 88  "' dP' 
Yb      Yb   dP 8I  dY 88""  d888888   dP'  
 YboodP  YbodP  8888Y" 888888    88  .d8888
 
code42cli version {}, by Code42 Software.
powered by py42 version {}.""".format(
    cliversion, py42version
)


class ArgumentParserError(Exception):
    pass


class CommandParser(argparse.ArgumentParser):
    def __init__(self, **kwargs):
        # noinspection PyTypeChecker
        super(CommandParser, self).__init__(
            formatter_class=RawDescriptionHelpFormatter, add_help=False, **kwargs
        )
        self.add_argument(
            "-h",
            "--help",
            action="help",
            default=argparse.SUPPRESS,
            help="Show this help message and exit.",
        )

    def prepare_command(self, command, path_parts):
        parser = self._get_parser(command, path_parts)
        self._load_argparse_config(command, parser)
        parser.set_defaults(func=lambda args: command(args, help_func=parser.print_help))
        return parser

    def prepare_cli_help(self, top_command):
        top_command.load_subcommands()
        self.description = _get_group_help(top_command)
        self.usage = SUPPRESS
        self.set_defaults(func=lambda _: self.print_help())
        return self

    def error(self, message):
        # overrides the behavior of when an error occurs when
        # arguments can't be successfully parsed. CommandInvoker catches this.
        raise ArgumentParserError(message)

    def _load_argparse_config(self, command, command_parser):
        arg_configs = command.get_arg_configs()
        required_group = command_parser.add_argument_group(u"required arguments")
        for arg in arg_configs:
            _add_argument(command_parser, arg_configs[arg].settings, required_group)

    def _get_parser(self, command, path_parts):
        usage = command.usage or SUPPRESS
        command.load_subcommands()
        description = _get_group_help(command) if command.subcommands else command.description
        subparser = self._get_subparser(path_parts)
        return subparser.add_parser(command.name, description=description, usage=usage)

    def _get_subparser(self, path_parts):
        global_subparser = self.add_subparsers()
        global_subparser.required = True
        subparsers = {(): global_subparser}
        parent_subparser = global_subparser

        # build out the entire path of subparsers up to the command
        for part in range(0, len(path_parts)):
            parent_path_parts = tuple(path_parts[:part])
            parent_subparser = subparsers.get(parent_path_parts)
            if not parent_subparser:
                parent_subparser = _get_parent_subparser(path_parts, part, subparsers)
            subparsers[parent_path_parts] = parent_subparser
        return parent_subparser


def _get_parent_subparser(path_parts, part, subparsers):
    grandparent_path_parts = tuple(path_parts[: part - 1])
    grandparent_subparser = subparsers[grandparent_path_parts]

    new_path = path_parts[part - 1]
    new_parser = grandparent_subparser.add_parser(new_path)
    parent_subparser = new_parser.add_subparsers()
    parent_subparser.required = True

    return parent_subparser


def _add_argument(parser, arg_settings, required_group):
    # register the settings of an ArgConfig object to an argparse parser
    options_list = arg_settings.pop(u"options_list")
    arg_settings = {key: arg_settings[key] for key in arg_settings if arg_settings[key] is not None}
    if arg_settings.get(u"required"):
        parser = required_group
    parser.add_argument(*options_list, **arg_settings)


def _get_group_help(command):
    descriptions = _build_group_command_descriptions(command)
    output = []
    name = command.name
    if not name:
        name = u"code42"
        output.append(BANNER)

    output.extend([u" \nAvailable commands in <{}>:".format(name), descriptions])
    return u"\n".join(output)


def _build_group_command_descriptions(command):
    subs = command.subcommands
    name_width = len(max([cmd.name for cmd in subs], key=len))
    lines = [u"  {} - {}".format(cmd.name.ljust(name_width), cmd.description) for cmd in subs]
    return u"\n".join(lines)



from py42.exceptions import Py42BadRequestError

from code42cli.cmds.detectionlists import DetectionListSubcommandLoader
from code42cli.commands import Command
from code42cli.cmds.detectionlists import (
    DetectionList,
    DetectionListHandlers,
    load_user_descriptions,
    update_user,
    try_handle_user_already_added_error,
    add_risk_tags,
    remove_risk_tags,
    load_username_description,
    handle_list_args,
)
from code42cli.util import get_user_id
from code42cli.cmds.detectionlists.enums import DetectionLists, DetectionListUserKeys, RiskTags


class HighRiskEmployeeSubcommandLoader(DetectionListSubcommandLoader):
    def __init__(self, root_command_name):
        super(HighRiskEmployeeSubcommandLoader, self).__init__(root_command_name)
        handlers = _create_handlers()
        self.detection_list = DetectionList.create_high_risk_employee_list(handlers)
        self._cmd_loader = self.detection_list.subcommand_loader

    def load_commands(self):
        cmds = self.detection_list.load_subcommands()
        cmds.extend(
            [
                Command(
                    u"add-risk-tags",
                    u"Associates risk tags with a user.",
                    u"code42 high-risk-employee add-risk-tags --username <username> --tag <risk-tags>",
                    handler=add_risk_tags,
                    arg_customizer=load_risk_tag_mgmt_descriptions,
                ),
                Command(
                    u"remove-risk-tags",
                    u"Disassociates risk tags from a user.",
                    u"code42 high-risk-employee remove-risk-tags --username <username> --tag <risk-tags>",
                    handler=remove_risk_tags,
                    arg_customizer=load_risk_tag_mgmt_descriptions,
                ),
            ]
        )
        return cmds


def _create_handlers():
    return DetectionListHandlers(
        add=add_high_risk_employee, remove=remove_high_risk_employee, load_add=_load_add_description
    )


def add_high_risk_employee(sdk, profile, username, cloud_alias=None, risk_tag=None, notes=None):
    """Adds an employee to the high risk employee detection list.
    
    Args:
        sdk (py42.sdk.SDKClient): py42.
        profile (C42Profile): Your code42 profile.
        username (str): The username of the employee to add.
        cloud_alias (str): An alternative email address for another cloud service.
        risk_tag (iter[str]): Risk tags associated with the employee.
        notes: (str): Notes about the employee.
    """
    risk_tag = handle_list_args(risk_tag)
    user_id = get_user_id(sdk, username)

    try:
        sdk.detectionlists.high_risk_employee.add(user_id)
        update_user(sdk, user_id, cloud_alias, risk_tag, notes)
    except Py42BadRequestError as err:
        list_name = u"{} list".format(DetectionLists.HIGH_RISK_EMPLOYEE)
        try_handle_user_already_added_error(err, username, list_name)
        raise


def remove_high_risk_employee(sdk, profile, username):
    """Removes an employee from the high risk employee detection list.
    
    Args:
        sdk (py42.sdk.SDKClient): py42.
        profile (C42Profile): Your code42 profile.
        username (str): The username of the employee to remove.
    """
    user_id = get_user_id(sdk, username)
    sdk.detectionlists.high_risk_employee.remove(user_id)


def _load_add_description(argument_collection):
    load_user_descriptions(argument_collection)
    load_risk_tag_description(argument_collection)


def load_risk_tag_description(argument_collection):
    risk_tag = (
        argument_collection.arg_configs.get(DetectionListUserKeys.RISK_TAG)
        or argument_collection.arg_configs[u"tag"]
    )
    risk_tag.as_multi_val_param()
    tags = u", ".join(list(RiskTags()))
    risk_tag.set_help(
        u"Risk tags associated with the employee. Options include: [{}].".format(tags)
    )


def load_risk_tag_mgmt_descriptions(argument_collection):
    load_username_description(argument_collection)
    load_risk_tag_description(argument_collection)



from code42cli.bulk import BulkCommandType


class HighRiskBulkCommandType(BulkCommandType):
    ADD_RISK_TAG = u"add-risk-tags"
    REMOVE_RISK_TAG = u"remove-risk-tags"

    def __iter__(self):
        parent_items = list(super(HighRiskBulkCommandType, self).__iter__())
        return iter([parent_items[0], parent_items[1], self.ADD_RISK_TAG, self.REMOVE_RISK_TAG])


class BulkDetectionList(object):
    def __init__(self):
        self.type = BulkCommandType

    def get_handler(self, handlers, cmd):
        handler = None
        if cmd == self.type.ADD:
            handler = handlers.add_employee
        elif cmd == self.type.REMOVE:
            handler = handlers.remove_employee
        return handler


class BulkHighRiskEmployee(BulkDetectionList):
    def __init__(self):
        super(BulkHighRiskEmployee, self).__init__()
        self.type = HighRiskBulkCommandType

    def get_handler(self, handlers, cmd):
        handler = super(BulkHighRiskEmployee, self).get_handler(handlers, cmd)
        if not handler:
            if cmd == self.type.ADD_RISK_TAG:
                handler = handlers.add_risk_tags
            elif cmd == self.type.REMOVE_RISK_TAG:
                handler = handlers.remove_risk_tags

        return handler



import platform
import signal
import sys

from py42.settings import set_user_agent_suffix

from code42cli import PRODUCT_NAME
from code42cli.cmds.detectionlists import departing_employee as de
from code42cli.cmds.detectionlists import high_risk_employee as hre
from code42cli.cmds.detectionlists.enums import DetectionLists
from code42cli.cmds.securitydata import main as secmain
from code42cli.cmds.alerts import main as alertmain
from code42cli.cmds.alerts.rules import commands as alertrules
from code42cli.cmds.legal_hold import commands as legalhold
from code42cli.cmds.profile import ProfileSubcommandLoader
from code42cli.commands import Command, SubcommandLoader
from code42cli.invoker import CommandInvoker


# Handle KeyboardInterrupts by just exiting instead of printing out a stack
def exit_on_interrupt(signal, frame):
    sys.exit(1)


signal.signal(signal.SIGINT, exit_on_interrupt)


# If on Windows, configure console session to handle ANSI escape sequences correctly
# source: https://bugs.python.org/issue29059
if platform.system().lower() == u"windows":
    from ctypes import windll, c_int, byref

    stdout_handle = windll.kernel32.GetStdHandle(c_int(-11))
    mode = c_int(0)
    windll.kernel32.GetConsoleMode(c_int(stdout_handle), byref(mode))
    mode = c_int(mode.value | 4)
    windll.kernel32.SetConsoleMode(c_int(stdout_handle), mode)


# Sets part of the user agent string that py42 attaches to requests for the purposes of
# identifying CLI users.
set_user_agent_suffix(PRODUCT_NAME)


class MainSubcommandLoader(SubcommandLoader):
    PROFILE = u"profile"
    SECURITY_DATA = u"security-data"
    ALERTS = u"alerts"
    ALERT_RULES = u"alert-rules"
    DEPARTING_EMPLOYEE = DetectionLists.DEPARTING_EMPLOYEE
    HIGH_RISK_EMPLOYEE = DetectionLists.HIGH_RISK_EMPLOYEE
    LEGAL_HOLD = u"legal-hold"

    def load_commands(self):
        detection_lists_description = (
            u"For adding and removing employees from the {} detection list."
        )
        return [
            Command(
                self.PROFILE,
                u"For managing Code42 settings.",
                subcommand_loader=self._create_profile_loader(),
            ),
            Command(
                self.SECURITY_DATA,
                u"Tools for getting security related data, such as file events.",
                subcommand_loader=self._create_security_data_loader(),
            ),
            Command(
                self.ALERTS,
                u"Tools for getting alert data.",
                subcommand_loader=self._create_alerts_loader(),
            ),
            Command(
                self.ALERT_RULES,
                u"Manage alert rules.",
                subcommand_loader=self._create_alert_rules_loader(),
            ),
            Command(
                self.DEPARTING_EMPLOYEE,
                detection_lists_description.format(u"departing employee"),
                subcommand_loader=self._create_departing_employee_loader(),
            ),
            Command(
                self.HIGH_RISK_EMPLOYEE,
                detection_lists_description.format(u"high risk employee"),
                subcommand_loader=self._create_high_risk_employee_loader(),
            ),
            Command(
                self.LEGAL_HOLD,
                u"For adding and removing employees to legal hold matters.",
                subcommand_loader=self._create_legal_hold_loader(),
            ),
        ]

    def _create_profile_loader(self):
        return ProfileSubcommandLoader(self.PROFILE)

    def _create_security_data_loader(self):
        return secmain.SecurityDataSubcommandLoader(self.SECURITY_DATA)

    def _create_alerts_loader(self):
        return alertmain.MainAlertsSubcommandLoader(self.ALERTS)

    def _create_alert_rules_loader(self):
        return alertrules.AlertRulesSubcommandLoader(self.ALERT_RULES)

    def _create_departing_employee_loader(self):
        return de.DepartingEmployeeSubcommandLoader(self.DEPARTING_EMPLOYEE)

    def _create_high_risk_employee_loader(self):
        return hre.HighRiskEmployeeSubcommandLoader(self.HIGH_RISK_EMPLOYEE)

    def _create_legal_hold_loader(self):
        return legalhold.LegalHoldSubcommandLoader(self.LEGAL_HOLD)


def main():
    top = Command(u"", u"", subcommand_loader=MainSubcommandLoader(u""))
    invoker = CommandInvoker(top)
    invoker.run(sys.argv[1:])


if __name__ == u"__main__":
    main()



from code42cli.compat import str
import code42cli.password as password
from code42cli.cmds.search_shared.cursor_store import get_all_cursor_stores_for_profile
from code42cli.config import ConfigAccessor, config_accessor, NoConfigProfileError
from code42cli.logger import get_main_cli_logger


class Code42Profile(object):
    def __init__(self, profile):
        self._profile = profile

    @property
    def name(self):
        return self._profile.name

    @property
    def authority_url(self):
        return self._profile[ConfigAccessor.AUTHORITY_KEY]

    @property
    def username(self):
        return self._profile[ConfigAccessor.USERNAME_KEY]

    @property
    def ignore_ssl_errors(self):
        return self._profile[ConfigAccessor.IGNORE_SSL_ERRORS_KEY]

    @property
    def has_stored_password(self):
        stored_password = password.get_stored_password(self)
        return stored_password is not None and stored_password != u""

    def get_password(self):
        pwd = password.get_stored_password(self)
        if not pwd:
            pwd = password.get_password_from_prompt()
        return pwd

    def __str__(self):
        return u"{0}: Username={1}, Authority URL={2}".format(
            self.name, self.username, self.authority_url
        )


def _get_profile(profile_name=None):
    """Returns the profile for the given name."""
    config_profile = config_accessor.get_profile(profile_name)
    return Code42Profile(config_profile)


def get_profile(profile_name=None):
    if profile_name is None:
        validate_default_profile()
    try:
        return _get_profile(profile_name)
    except NoConfigProfileError as ex:
        logger = get_main_cli_logger()
        logger.print_and_log_error(str(ex))
        _print_create_profile_help()
        exit(1)


def default_profile_exists():
    try:
        profile = _get_profile()
        return profile.name and profile.name != ConfigAccessor.DEFAULT_VALUE
    except NoConfigProfileError:
        return False


def is_default_profile(name):
    if default_profile_exists():
        default = get_profile()
        return name == default.name


def validate_default_profile():
    if not default_profile_exists():
        existing_profiles = get_all_profiles()
        if not existing_profiles:
            print_and_log_no_existing_profile()
        else:
            _print_set_default_profile_help(existing_profiles)
        exit(1)


def profile_exists(profile_name=None):
    try:
        _get_profile(profile_name)
        return True
    except NoConfigProfileError:
        return False


def switch_default_profile(profile_name):
    profile = get_profile(profile_name)  # Handles if profile does not exist.
    config_accessor.switch_default_profile(profile.name)


def create_profile(name, server, username, ignore_ssl_errors):
    if profile_exists(name):
        logger = get_main_cli_logger()
        logger.print_and_log_error(u"A profile named '{}' already exists.".format(name))
        exit(1)

    config_accessor.create_profile(name, server, username, ignore_ssl_errors)


def delete_profile(profile_name):
    profile = _get_profile(profile_name)
    if password.get_stored_password(profile) is not None:
        password.delete_password(profile)
    cursor_stores = get_all_cursor_stores_for_profile(profile_name)
    for store in cursor_stores:
        store.clean()
    config_accessor.delete_profile(profile_name)
    get_main_cli_logger().print_info(u"Profile '{}' has been deleted.".format(profile_name))


def update_profile(name, server, username, ignore_ssl_errors):
    config_accessor.update_profile(name, server, username, ignore_ssl_errors)


def get_all_profiles():
    profiles = [Code42Profile(profile) for profile in config_accessor.get_all_profiles()]
    return profiles


def get_stored_password(profile_name=None):
    profile = get_profile(profile_name)
    return password.get_stored_password(profile)


def set_password(new_password, profile_name=None):
    profile = get_profile(profile_name)
    password.set_password(profile, new_password)


def print_and_log_no_existing_profile():
    logger = get_main_cli_logger()
    logger.print_and_log_error(u"No existing profile.")
    _print_create_profile_help()


def _print_create_profile_help():
    logger = get_main_cli_logger()
    logger.print_info(u"\nTo add a profile, use: ")
    logger.print_bold(u"\tcode42 profile create <profile-name> <authority-URL> <username>\n")


def _print_set_default_profile_help(existing_profiles):
    logger = get_main_cli_logger()
    logger.print_info(
        u"\nNo default profile set.\n"
        u"\nUse the --profile flag to specify which profile to use.\n"
        u"\nTo set the default profile (used whenever --profile argument is not provided), use:"
    )
    logger.print_bold(u"\tcode42 profile use <profile-name>")
    logger.print_info(u"\nExisting profiles:")
    for profile in existing_profiles:
        logger.print_info("\t{}".format(profile))
    logger.print_info(u"")



from __future__ import print_function
import sys
import shutil

from collections import OrderedDict
from functools import wraps
from os import makedirs, path
from signal import signal, getsignal, SIGINT

from code42cli.compat import open, str
from code42cli.errors import UserDoesNotExistError

_PADDING_SIZE = 3


def get_input(prompt):
    """Uses correct input function based on Python version."""
    # pylint: disable=undefined-variable
    if sys.version_info >= (3, 0):
        return input(prompt)
    else:
        return raw_input(prompt)


def does_user_agree(prompt):
    """Prompts the user and checks if they said yes."""
    ans = get_input(prompt)
    ans = ans.strip().lower()
    return ans == u"y"


def get_user_project_path(subdir=u""):
    """The path on your user dir to /.code42cli/[subdir]."""
    package_name = __name__.split(u".")[0]
    home = path.expanduser(u"~")
    hidden_package_name = u".{0}".format(package_name)
    user_project_path = path.join(home, hidden_package_name, subdir)
    if not path.exists(user_project_path):
        makedirs(user_project_path)
    return user_project_path


def open_file(file_path, mode, action):
    """Wrapper for opening files, useful for testing purposes."""
    with open(file_path, mode, encoding=u"utf-8") as f:
        return action(f)


def is_interactive():
    return sys.stdin.isatty()


def get_url_parts(url_str):
    parts = url_str.split(u":")
    port = None
    if len(parts) > 1 and parts[1] != u"":
        port = int(parts[1])
    return parts[0], port


def find_format_width(record, header):
    """Fetches needed keys/items to be displayed based on header keys.
    
    Finds the largest string against each column so as to decide the padding size for the column.
    
    Args:
        record (list of dict), data to be formatted.  
        header (dict), key-value where keys should map to keys of record dict and
          value is the corresponding column name to be displayed on the cli.
    
    Returns:
        tuple (list of dict, dict), i.e Filtered records, padding size of columns.
    """
    rows = [header]

    # Set default max width items to column names
    max_width_item = dict(header.items())
    for record_row in record:
        row = OrderedDict()
        for header_key in header.keys():
            row[header_key] = record_row[header_key]
            max_width_item[header_key] = max(
                max_width_item[header_key], str(record_row[header_key]), key=len
            )
        rows.append(row)
    column_size = {key: len(value) for key, value in max_width_item.items()}
    return rows, column_size


def format_to_table(rows, column_size):
    """Prints result in left justified format in a tabular form."""
    for row in rows:
        for key in row.keys():
            print(str(row[key]).ljust(column_size[key] + _PADDING_SIZE), end=u" ")
        print(u"")


def format_string_list_to_columns(string_list, max_width=None):
    """Prints a list of strings in justified columns and fits them neatly into specified width."""
    if not string_list:
        return
    if not max_width:
        max_width, _ = shutil.get_terminal_size()
    column_width = len(max(string_list, key=len)) + _PADDING_SIZE
    num_columns = int(max_width / column_width)
    format_string = u"{{:<{0}}}".format(column_width) * num_columns
    batches = [string_list[i : i + num_columns] for i in range(0, len(string_list), num_columns)]
    padding = [u"" for _ in range(num_columns)]
    for batch in batches:
        print(format_string.format(*batch + padding))
    print()


def color_text_red(text):
    return u"\033[91m{}\033[0m".format(text)


class warn_interrupt(object):
    """A context decorator class used to wrap functions where a keyboard interrupt could potentially
    leave things in a bad state. Warns the user with provided message and exits when wrapped 
    function is complete. Requires user to ctrl-c a second time to force exit.
    
    Usage:
    
    @warn_interrupt(warning="example message")
    def my_important_func():
        pass
    """

    def __init__(self, warning="Cancelling operation cleanly, one moment... "):
        self.warning = warning
        self.old_handler = None
        self.interrupted = False
        self.exit_instructions = "Hit CTRL-C again to force quit."

    def __enter__(self):
        self.old_handler = getsignal(SIGINT)
        signal(SIGINT, self._handle_interrupts)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.interrupted:
            exit(1)
        signal(SIGINT, self.old_handler)
        return False

    def _handle_interrupts(self, sig, frame):
        if not self.interrupted:
            self.interrupted = True
            print("\n{}\n{}".format(self.warning, self.exit_instructions), file=sys.stderr)
        else:
            exit()

    def __call__(self, func):
        @wraps(func)
        def inner(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return inner


def get_user_id(sdk, username):
    """Returns the user's UID (referred to by `user_id` in detection lists). Raises 
    `UserDoesNotExistError` if the user doesn't exist in the Code42 server.
    
    Args:
        sdk (py42.sdk.SDKClient): The py42 sdk.
        username (str or unicode): The username of the user to get an ID for.
    
    Returns:
         str: The user ID for the user with the given username.
    """
    users = sdk.users.get_by_username(username)[u"users"]
    if not users:
        raise UserDoesNotExistError(username)
    return users[0][u"userUid"]



import sys

import difflib

from py42.exceptions import Py42HTTPError, Py42ForbiddenError

from code42cli.compat import str
from code42cli.errors import Code42CLIError
from code42cli.parser import ArgumentParserError, CommandParser
from code42cli.logger import get_main_cli_logger

_DIFFLIB_CUT_OFF = 0.7


class CommandInvoker(object):

    _COMMAND_KEYWORDS = {}
    _COMMAND_ARG_KEYWORDS = {}

    def __init__(self, top_command, cmd_parser=None):
        self._top_command = top_command
        self._cmd_parser = cmd_parser or CommandParser()
        self._commands = {u"": self._top_command}

    def run(self, input_args):
        """Locates a command that matches the one specified by
        `input_args` and runs it with the supplied parameters.
        
        Args:
            input_args (iter[str]): the full list of arguments
            supplied by the user to `code42` cli command.
        """
        invocation_str = u"code42 {}".format(u" ".join(input_args))
        try:
            path_parts = self._get_path_parts(input_args)
            command = self._commands.get(u" ".join(path_parts))
            self._try_run_command(command, path_parts, input_args)
        except Code42CLIError as err:
            logger = get_main_cli_logger()
            logger.print_and_log_error(str(err))
        except Py42ForbiddenError as err:
            logger = get_main_cli_logger()
            logger.log_verbose_error(invocation_str, err.response.request)
            logger.print_and_log_permissions_error()
            logger.print_errors_occurred_message()
        except Py42HTTPError as err:
            logger = get_main_cli_logger()
            logger.log_verbose_error(invocation_str, err.response.request)
            logger.print_errors_occurred_message()
        except Exception:
            logger = get_main_cli_logger()
            logger.log_verbose_error(invocation_str)
            logger.print_errors_occurred_message()

    def _get_path_parts(self, input_args):
        """Gets the portion of `input_args` that refers to a
        valid command or subcommand, removing parameters.
        For example, `input_args` of ["command", "sub", "--arg", "argval"]
        would return ["command", "sub"], assuming "command" is a top level command,
        "sub" is a subcommand of "command", and the rest of the values are normal parameters.
        Returns an empty string if a valid command or subcommand is not found.
        """
        path = u""
        node = self._commands[u""]
        # step through each segment of input_args until we find
        # something that _isn't_ a command or subcommand.
        for arg in input_args:
            new_path = u"{} {}".format(path, arg).strip()
            self._load_subcommands(path, node)
            node = self._commands.get(new_path)
            if not node:
                break
            path = new_path
        return path.split()

    def _load_subcommands(self, path, node):
        """Discovers a command's subcommands and registers them
        to the available list of commands for this Invoker."""
        node.load_subcommands()
        for command in node.subcommands:
            new_key = u"{} {}".format(path, command.name).strip()
            self._commands[new_key] = command
            self._set_command_keywords(new_key)

    def _try_run_command(self, command, path_parts, input_args):
        """Runs a command called using `path_parts` by parsing
        `input_args` and calling the command's handler."""
        parser = None
        try:
            if not path_parts:
                parser = self._cmd_parser.prepare_cli_help(command)
            else:
                parser = self._cmd_parser.prepare_command(command, path_parts)
                self._set_argument_keywords(path_parts[0], command.get_arg_configs())
            parsed_args = self._cmd_parser.parse_args(input_args)
            parsed_args.func(parsed_args)
        except ArgumentParserError as err:
            logger = get_main_cli_logger()
            logger.print_and_log_error(u"{}".format(err))
            possible_correct_words = self._find_incorrect_word_match(err, path_parts)
            if possible_correct_words:
                logger.print_info(u"Did you mean one of the following?")
                for possible_correct_word in possible_correct_words:
                    logger.print_info(u"     {}".format(possible_correct_word))

            else:
                parser.print_help(sys.stderr)
            sys.exit(2)

    @staticmethod
    def _get_arg_flags(arguments):
        flag_names = []
        for arg in arguments.values():
            arg_flags = [name for name in arg.settings["options_list"] if name.startswith("-")]
            flag_names.extend(arg_flags)
        return flag_names

    def _set_argument_keywords(self, command_key, arguments):
        self._COMMAND_ARG_KEYWORDS[command_key] = set()
        self._COMMAND_ARG_KEYWORDS[command_key].update(CommandInvoker._get_arg_flags(arguments))

    def _set_command_keywords(self, new_key):
        """Creates a dictionary, with top level command as key and set of all its subcommands
        as values.
        """
        command_keys = new_key.split()
        if len(command_keys) == 1:
            self._COMMAND_KEYWORDS[command_keys[0]] = set()
        else:
            self._COMMAND_KEYWORDS[command_keys[0]].update(command_keys[1:])

    def _find_incorrect_word_match(self, error, path_parts):
        possible_correct_words = []

        try:
            # Here we assume the error string contains ":", for case where it doesn't we
            # assume the error is not due to misspelled word and we return error as is.
            error_detail, unmatched_words = str(error).split(u":")
        except ValueError:
            return possible_correct_words

        if not unmatched_words or error_detail != u"unrecognized arguments":
            return possible_correct_words

        # Arg-parser sets the first/leftmost incorrect command keyword in the error message.
        unmatched_word = unmatched_words.split()[0]

        if not path_parts:
            available_values = self._COMMAND_KEYWORDS.keys()
        elif unmatched_word.strip().startswith("-"):
            available_values = self._COMMAND_ARG_KEYWORDS[path_parts[0]]
        else:
            available_values = self._COMMAND_KEYWORDS[path_parts[0]]

        return difflib.get_close_matches(unmatched_word, available_values, cutoff=_DIFFLIB_CUT_OFF)



ERRORED = False


_FORMAT_VALUE_ERROR_MESSAGE = (
    u"input must be a date/time string (e.g. 'yyyy-MM-dd', "
    u"'yy-MM-dd HH:MM', 'yy-MM-dd HH:MM:SS'), or a short value in days, "
    u"hours, or minutes (e.g. 30d, 24h, 15m)"
)


class BadFileError(Exception):
    def __init__(self, file_path, *args, **kwargs):
        self.file_path = file_path
        super(BadFileError, self).__init__()


class EmptyFileError(BadFileError):
    def __init__(self, file_path):
        super(EmptyFileError, self).__init__(file_path, u"Given empty file {}.".format(file_path))


class Code42CLIError(Exception):
    pass


class UserAlreadyAddedError(Code42CLIError):
    def __init__(self, username, list_name):
        msg = u"'{}' is already on the {}.".format(username, list_name)
        super(UserAlreadyAddedError, self).__init__(msg)


class UnknownRiskTagError(Code42CLIError):
    def __init__(self, bad_tags):
        tags = u", ".join(bad_tags)
        super(UnknownRiskTagError, self).__init__(
            u"The following risk tags are unknown: '{}'.".format(tags)
        )


class InvalidRuleTypeError(Code42CLIError):
    def __init__(self, rule_id, source):
        msg = u"Only alert rules with a source of 'Alerting' can be targeted by this command. "
        msg += "Rule {0} has a source of '{1}'."
        super(InvalidRuleTypeError, self).__init__(msg.format(rule_id, source))


class UserDoesNotExistError(Code42CLIError):
    """An error to represent a username that is not in our system. The CLI shows this error when 
    the user tries to add or remove a user that does not exist. This error is not shown during 
    bulk add or remove."""

    def __init__(self, username):
        super(UserDoesNotExistError, self).__init__(u"User '{}' does not exist.".format(username))


class UserNotInLegalHoldError(Code42CLIError):
    def __init__(self, username, matter_id):
        super(UserNotInLegalHoldError, self).__init__(
            u"User '{}' is not an active member of legal hold matter '{}'".format(
                username, matter_id
            )
        )


class LegalHoldNotFoundOrPermissionDeniedError(Code42CLIError):
    def __init__(self, matter_id):
        super(LegalHoldNotFoundOrPermissionDeniedError, self).__init__(
            u"Matter with id={} either does not exist or your profile does not have permission to "
            u"view it.".format(matter_id)
        )


class DateArgumentError(Code42CLIError):
    def __init__(self, message=_FORMAT_VALUE_ERROR_MESSAGE):
        super(DateArgumentError, self).__init__(message)



import csv

from code42cli.errors import BadFileError


class CliFileReader(object):
    _ROWS_COUNT = -1

    def __init__(self, file_path):
        self.file_path = file_path

    def __call__(self, *args, **kwargs):
        pass

    def get_rows_count(self):
        if self._ROWS_COUNT == -1:
            self._ROWS_COUNT = sum(1 for _ in open(self.file_path))
        if self._ROWS_COUNT == 0:
            raise BadFileError(u"Given empty file {}.".format(self.file_path))
        return self._ROWS_COUNT


class CSVReader(CliFileReader):
    """A generator that yields header keys mapped to row values from a csv file."""

    def __init__(self, file_path):
        with open(file_path) as f:
            try:
                self.has_header = csv.Sniffer().has_header(next(f))
            except StopIteration:
                raise BadFileError(u"Given empty file {}.".format(file_path))
        super(CSVReader, self).__init__(file_path)

    def __call__(self, *args, **kwargs):
        for row in csv.DictReader(kwargs.get(u"bulk_file")):
            yield row

    def get_rows_count(self):
        rows_count = super(CSVReader, self).get_rows_count()
        return rows_count - 1 if self.has_header else rows_count


class FlatFileReader(CliFileReader):
    """A generator that yields a single-value per row from a file."""

    def __call__(self, *args, **kwargs):
        for row in kwargs[u"bulk_file"]:
            yield row


def create_csv_reader(file_path):
    return CSVReader(file_path)


def create_flat_file_reader(file_path):
    return FlatFileReader(file_path)



import os

from code42cli.args import ArgConfig
from code42cli.commands import Command, SubcommandLoader
from code42cli.bulk import generate_template, BulkCommandType
from code42cli.cmds.legal_hold import (
    add_user,
    remove_user,
    get_matters,
    add_bulk_users,
    remove_bulk_users,
    show_matter,
)


class LegalHoldSubcommandLoader(SubcommandLoader):
    ADD_USER = "add-user"
    REMOVE_USER = "remove-user"
    LIST = "list"
    SHOW = "show"
    BULK = "bulk"

    def __init__(self, root_command_name):
        super(LegalHoldSubcommandLoader, self).__init__(root_command_name)
        self._bulk_subcommand_loader = LegalHoldBulkSubcommandLoader(self.BULK)

    def load_commands(self):
        """Sets up the `legal-hold` subcommand with all of its subcommands."""
        usage_prefix = u"code42 legal-hold"

        add = Command(
            self.ADD_USER,
            u"Add a user to a legal hold matter.",
            u"{} add-user --matter-id <id> --username <username>".format(usage_prefix),
            handler=add_user,
            arg_customizer=_customize_add_arguments,
        )

        remove = Command(
            self.REMOVE_USER,
            u"Remove a user from a legal hold matter.",
            u"{} remove-user --matter-id <id> --username <username>".format(usage_prefix),
            handler=remove_user,
            arg_customizer=_customize_remove_arguments,
        )

        list_matters = Command(
            self.LIST,
            u"Fetch existing legal hold matters.",
            u"{} list".format(usage_prefix),
            handler=get_matters,
        )

        show = Command(
            self.SHOW,
            u"Fetch all legal hold custodians for a given matter.",
            u"{} show <matter-id>".format(usage_prefix),
            handler=show_matter,
            arg_customizer=_customize_show_arguments,
        )

        bulk = Command(
            self.BULK,
            u"Tools for executing bulk commands.",
            subcommand_loader=self._bulk_subcommand_loader,
        )

        return [add, remove, list_matters, show, bulk]


class LegalHoldBulkSubcommandLoader(SubcommandLoader):
    GENERATE_TEMPLATE = u"generate-template"
    ADD = u"add"
    REMOVE = u"remove"

    def load_commands(self):
        """Sets up the `legal-hold bulk` subcommands."""
        usage_prefix = u"code42 legal-hold bulk"

        bulk_add = Command(
            u"add-user",
            u"Bulk add users to legal hold matters from a csv file. CSV file format: matter_id,username",
            u"{} add-user <filename>".format(usage_prefix),
            handler=add_bulk_users,
            arg_customizer=_customize_bulk_arguments,
        )

        bulk_remove = Command(
            u"remove-user",
            u"Bulk remove users from legal hold matters from a csv file. CSV file format: matter_id,username",
            u"{} remove-user <filename>".format(usage_prefix),
            handler=remove_bulk_users,
            arg_customizer=_customize_bulk_arguments,
        )

        generate_template_cmd = Command(
            u"generate-template",
            u"Generate the necessary csv template needed for bulk adding users.",
            u"{} generate-template <cmd> <optional args>".format(usage_prefix),
            handler=_generate_template_file,
            arg_customizer=_load_bulk_generate_template_description,
        )

        return [bulk_add, bulk_remove, generate_template_cmd]


def _customize_add_arguments(argument_collection):
    matter = argument_collection.arg_configs["matter_id"]
    matter.add_short_option_name("-m")
    matter.set_help("ID of the legal hold matter user will be added to. Required.")
    username = argument_collection.arg_configs["username"]
    username.add_short_option_name("-u")
    username.set_help("The username of the user to add to the matter. Required.")


def _customize_remove_arguments(argument_collection):
    matter = argument_collection.arg_configs["matter_id"]
    matter.add_short_option_name("-m")
    matter.set_help("ID of the legal hold matter user will be removed from. Required.")
    username = argument_collection.arg_configs["username"]
    username.add_short_option_name("-u")
    username.set_help("The username of the user to remove from the matter. Required.")


def _customize_show_arguments(argument_collection):
    matter_id = argument_collection.arg_configs[u"matter_id"]
    matter_id.set_help(u"ID of the legal hold matter.")
    args = {
        u"include_inactive": ArgConfig(
            u"--include-inactive",
            action=u"store_true",
            help=u"Include list of users who are no longer actively on this matter.",
        ),
        u"include_policy": ArgConfig(
            u"--include-policy",
            action=u"store_true",
            help=u"Include the preservation policy (in json format) for this matter.",
        ),
    }
    argument_collection.extend(args)


def _customize_bulk_arguments(argument_collection):
    file_name = argument_collection.arg_configs[u"file_name"]
    file_name.set_help(
        u"The path to the csv file with columns 'matter_id,username' "
        u"for bulk adding users to legal hold."
    )


def _generate_template_file(cmd, path=None):
    handler = None
    filename = u"legal_hold.csv"
    if cmd == BulkCommandType.ADD:
        handler = add_user
        filename = u"add_users_to_{}".format(filename)
    elif cmd == BulkCommandType.REMOVE:
        handler = remove_user
        filename = u"remove_users_from_{}".format(filename)
    if not path:
        path = os.path.join(os.getcwd(), filename)
    generate_template(handler, path)


def _load_bulk_generate_template_description(argument_collection):
    cmd_type = argument_collection.arg_configs[u"cmd"]
    cmd_type.set_help(u"The type of command the template will be used for.")
    cmd_type.set_choices(BulkCommandType())



from codecs import open
from os import path
from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

about = {}
with open(path.join(here, "src", "code42cli", "__version__.py"), encoding="utf8") as fh:
    exec(fh.read(), about)

with open(path.join(here, "README.md"), "r", "utf-8") as f:
    readme = f.read()

setup(
    name="code42cli",
    version=about["__version__"],
    description="The official command line tool for interacting with Code42",
    long_description=readme,
    long_description_content_type="text/markdown",
    packages=find_packages("src"),
    package_dir={"": "src"},
    python_requires=">3, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, <4",
    install_requires=[
        "c42eventextractor==0.3.1",
        "keyring==18.0.1",
        "keyrings.alt==3.2.0",
        "py42>=1.2.0",
    ],
    license="MIT",
    include_package_data=True,
    zip_safe=False,
    extras_require={
        "dev": [
            "pre-commit",
            "pytest==4.6.5",
            "pytest-cov == 2.8.1",
            "pytest-mock==2.0.0",
            "recommonmark",
            "sphinx",
            "sphinx_rtd_theme",
            "tox==3.14.3",
        ]
    },
    classifiers=[
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    scripts=["bin/code42cli_completer"],
    entry_points={"console_scripts": ["code42=code42cli.main:main"]},
)



from c42eventextractor.extractors import AlertExtractor
from py42.sdk.queries.alerts.filters import (
    Actor,
    AlertState,
    Severity,
    DateObserved,
    Description,
    RuleName,
    RuleId,
    RuleType,
)

import code42cli.cmds.search_shared.enums as enums
import code42cli.errors as errors
from code42cli.cmds.search_shared.cursor_store import AlertCursorStore
from code42cli.cmds.search_shared.extraction import (
    verify_begin_date_requirements,
    create_handlers,
    exit_if_advanced_query_used_with_other_search_args,
    create_time_range_filter,
)
from code42cli.logger import get_main_cli_logger

logger = get_main_cli_logger()


def extract(sdk, profile, output_logger, args):
    """Extracts alerts using the given command-line arguments.

        Args:
            sdk (py42.sdk.SDKClient): The py42 sdk.
            profile (Code42Profile): The profile under which to execute this command.
            output_logger (Logger): The logger specified by which subcommand you use. For example,
                print: uses a logger that streams to stdout.
                write-to: uses a logger that logs to a file.
                send-to: uses a logger that sends logs to a server.
            args: Command line args used to build up alert query filters.
    """
    store = AlertCursorStore(profile.name) if args.incremental else None
    handlers = create_handlers(output_logger, store, event_key=u"alerts", sdk=sdk)
    extractor = AlertExtractor(sdk, handlers)
    if args.advanced_query:
        exit_if_advanced_query_used_with_other_search_args(args)
        extractor.extract_advanced(args.advanced_query)
    else:
        verify_begin_date_requirements(args, store)
        _verify_alert_state(args.state)
        _verify_alert_severity(args.severity)
        filters = _create_alert_filters(args)
        extractor.extract(*filters)
    if handlers.TOTAL_EVENTS == 0 and not errors.ERRORED:
        logger.print_info(u"No results found\n")


def _verify_alert_state(alert_state):
    options = list(enums.AlertState())
    if alert_state and alert_state not in options:
        logger.print_and_log_error(
            u"'{0}' is not a valid alert state, options are {1}.".format(alert_state, options)
        )
        exit(1)


def _verify_alert_severity(severity):
    if severity is None:
        return
    options = list(enums.AlertSeverity())
    for s in severity:
        if s not in options:
            logger.print_and_log_error(
                u"'{0}' is not a valid alert severity, options are {1}".format(s, options)
            )
            exit(1)


def _create_alert_filters(args):
    filters = []
    alert_timestamp_filter = create_time_range_filter(DateObserved, args.begin, args.end)
    not alert_timestamp_filter or filters.append(alert_timestamp_filter)
    not args.actor or filters.append(Actor.is_in(args.actor))
    not args.actor_contains or [filters.append(Actor.contains(arg)) for arg in args.actor_contains]
    not args.exclude_actor or filters.append(Actor.not_in(args.exclude_actor))
    not args.exclude_actor_contains or [
        filters.append(Actor.not_contains(arg)) for arg in args.exclude_actor_contains
    ]
    not args.rule_name or filters.append(RuleName.is_in(args.rule_name))
    not args.exclude_rule_name or filters.append(RuleName.not_in(args.exclude_rule_name))
    not args.rule_id or filters.append(RuleId.is_in(args.rule_id))
    not args.exclude_rule_id or filters.append(RuleId.not_in(args.exclude_rule_id))
    not args.rule_type or filters.append(RuleType.is_in(args.rule_type))
    not args.exclude_rule_type or filters.append(RuleType.not_in(args.exclude_rule_type))
    not args.description or filters.append(Description.contains(args.description))
    not args.severity or filters.append(Severity.is_in(args.severity))
    not args.state or filters.append(AlertState.eq(args.state))
    return filters



import os, logging, sys, traceback
from logging.handlers import RotatingFileHandler
from threading import Lock
import copy

from code42cli.compat import str
from code42cli.util import get_user_project_path, is_interactive, color_text_red


logger_deps_lock = Lock()
ERROR_LOG_FILE_NAME = u"code42_errors.log"
_PERMISSIONS_MESSAGE = (
    u"You do not have the necessary permissions to perform this task. "
    + u"Try using or creating a different profile."
)


def get_logger_for_stdout(name_suffix=u"main", formatter=None):
    logger = logging.getLogger(u"code42_stdout_{}".format(name_suffix))
    if logger_has_handlers(logger):
        return logger

    with logger_deps_lock:
        if not logger_has_handlers(logger):
            handler = logging.StreamHandler(sys.stdout)
            formatter = formatter or _get_standard_formatter()
            logger.setLevel(logging.INFO)
            return add_handler_to_logger(logger, handler, formatter)
    return logger


def _get_standard_formatter():
    return logging.Formatter(u"%(message)s")


def _get_error_log_path():
    log_path = get_user_project_path(u"log")
    return os.path.join(log_path, ERROR_LOG_FILE_NAME)


def _create_error_file_handler():
    log_path = _get_error_log_path()
    return RotatingFileHandler(log_path, maxBytes=250000000, encoding=u"utf-8", delay=True)


def add_handler_to_logger(logger, handler, formatter):
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


def logger_has_handlers(logger):
    return len(logger.handlers)


def _get_error_file_logger():
    """Gets the logger where raw exceptions are logged."""
    logger = logging.getLogger(u"code42_error_logger")
    if logger_has_handlers(logger):
        return logger

    with logger_deps_lock:
        if not logger_has_handlers(logger):
            formatter = _create_formatter_for_error_file()
            handler = _create_error_file_handler()
            return add_handler_to_logger(logger, handler, formatter)
    return logger


def get_view_exceptions_location_message():
    """Returns the error message that is printed when errors occur."""
    path = _get_error_log_path()
    return u"View exceptions that occurred at {}.".format(path)


def _get_user_error_logger():
    if is_interactive():
        return _get_interactive_user_error_logger()
    else:
        return _get_error_file_logger()


class RedStderrHandler(logging.StreamHandler):
    """Logging handler for logging error messages to stderr using red scary text prefixed by the 
    word `ERROR`. For logging info to stderr, it will not add the scary red text."""

    def __init__(self):
        super(RedStderrHandler, self).__init__(sys.stderr)

    def emit(self, record):
        if record.levelno == logging.ERROR:
            message = _get_red_error_text(record.msg)
            record = copy.copy(record)
            record.msg = message
        super(RedStderrHandler, self).emit(record)


def _get_interactive_user_error_logger():
    """This logger has two handlers, one for stderr and one for the error log file."""
    logger = logging.getLogger(u"code42_stderr_main")
    if logger_has_handlers(logger):
        return logger

    with logger_deps_lock:
        if not logger_has_handlers(logger):
            stderr_handler = RedStderrHandler()
            stderr_formatter = _get_standard_formatter()
            stderr_handler.setFormatter(stderr_formatter)

            file_handler = _create_error_file_handler()
            file_formatter = _create_formatter_for_error_file()
            file_handler.setFormatter(file_formatter)

            add_handler_to_logger(logger, stderr_handler, stderr_formatter)
            add_handler_to_logger(logger, file_handler, file_formatter)

            logger.setLevel(logging.INFO)
            return logger
    return logger


def _create_formatter_for_error_file():
    return logging.Formatter(u"%(asctime)s %(message)s")


def _get_red_error_text(text):
    return color_text_red(u"ERROR: {}".format(text))


def get_progress_logger(handler=None):
    logger = logging.getLogger(u"code42cli_progress_bar")
    if logger_has_handlers(logger):
        return logger

    with logger_deps_lock:
        if not logger_has_handlers(logger):
            handler = handler or InPlaceStreamHandler()
            formatter = _get_standard_formatter()
            logger.setLevel(logging.INFO)
            return add_handler_to_logger(logger, handler, formatter)
    return logger


class InPlaceStreamHandler(logging.StreamHandler):
    def __init__(self):
        super(InPlaceStreamHandler, self).__init__(sys.stdout)

    def emit(self, record):
        # Borrowed some from python3's logging.StreamHandler to make work on python2.
        try:
            msg = u"\r{}\r".format(self.format(record))
            stream = self.stream
            stream.write(msg)
            self.flush()
        except RuntimeError as err:
            if u"recursion" in str(err):
                raise
        except Exception:
            self.handleError(record)


class CliLogger(object):
    """There are three loggers part of the CliLogger. The following table illustrates where they 
    log to in both interactive mode and non-interactive mode.
    """

    def __init__(self):
        """The following properties explain how to log to different locations:
        
        `self._info_logger` is for when you want to display simple information, like 
            `profile list`. This does _not_ go to the log file.
            
        `self._user_error_logger` is for when you want to print in red text to the user. It also 
            goes to the log file for debugging purposes.
        
        `self._error_file_logger` logs directly to the error file and is only meant for verbose 
            debugging information, such as raw exceptions.
        """
        self._info_logger = get_logger_for_stdout()
        self._user_error_logger = _get_user_error_logger()
        self._error_file_logger = _get_error_file_logger()

    def print_info(self, message):
        self._info_logger.info(message)

    def print_bold(self, message):
        self._info_logger.info(u"\033[1m{}\033[0m".format(message))

    def print_and_log_error(self, message):
        """Logs red error text to stderr and non-color messages to the log file."""
        self._user_error_logger.error(message)

    def print_and_log_info(self, message):
        """Prints to stderr and the log file."""
        self._user_error_logger.info(message)

    def log_error(self, err):
        if err:
            message = str(err)  # Filter out empty string logs.
            if message:
                self._error_file_logger.error(message)

    def print_errors_occurred_message(self, additional_info=None):
        """Prints a message telling the user how to retrieve error logs."""
        locations_message = get_view_exceptions_location_message()
        message = (
            u"{}\n{}".format(additional_info, locations_message)
            if additional_info
            else locations_message
        )
        # Use `info()` because this message is pointless in the error log.
        self.print_info(_get_red_error_text(message))

    def log_verbose_error(self, invocation_str=None, http_request=None):
        """For logging traces, invocation strs, and request parameters during exceptions to the 
        error log file."""
        prefix = (
            u"Exception occurred."
            if not invocation_str
            else u"Exception occurred from input: '{}'.".format(invocation_str)
        )
        message = u"{}. See error below.".format(prefix)
        self.log_error(message)
        self.log_error(traceback.format_exc())
        if http_request:
            self.log_error(u"Request parameters: {}".format(http_request.body))

    def print_and_log_permissions_error(self):
        self.print_and_log_error(_PERMISSIONS_MESSAGE)

    def log_permissions_error(self):
        self.log_error(_PERMISSIONS_MESSAGE)


def get_main_cli_logger():
    return CliLogger()



from c42eventextractor.extractors import FileEventExtractor
from py42.sdk.queries.fileevents.filters import *

from code42cli.cmds.search_shared.enums import ExposureType as ExposureTypeOptions
from code42cli.cmds.search_shared.cursor_store import FileEventCursorStore
from code42cli.cmds.search_shared.extraction import (
    verify_begin_date_requirements,
    create_handlers,
    exit_if_advanced_query_used_with_other_search_args,
    create_time_range_filter,
)
import code42cli.errors as errors
from code42cli.logger import get_main_cli_logger

logger = get_main_cli_logger()


def extract(sdk, profile, output_logger, args):
    """Extracts file events using the given command-line arguments.

        Args:
            sdk (py42.sdk.SDKClient): The py42 sdk.
            profile (Code42Profile): The profile under which to execute this command.
            output_logger (Logger): The logger specified by which subcommand you use. For example,
                print: uses a logger that streams to stdout.
                write-to: uses a logger that logs to a file.
                send-to: uses a logger that sends logs to a server.
            args: Command line args used to build up file event query filters.
    """
    store = FileEventCursorStore(profile.name) if args.incremental else None
    handlers = create_handlers(output_logger, store, event_key=u"fileEvents")
    extractor = FileEventExtractor(sdk, handlers)
    if args.advanced_query:
        exit_if_advanced_query_used_with_other_search_args(args)
        extractor.extract_advanced(args.advanced_query)
    else:
        verify_begin_date_requirements(args, store)
        if args.type:
            _verify_exposure_types(args.type)
        filters = _create_file_event_filters(args)
        extractor.extract(*filters)
    if handlers.TOTAL_EVENTS == 0 and not errors.ERRORED:
        logger.print_info(u"No results found.")


def _verify_exposure_types(exposure_types):
    if exposure_types is None:
        return
    options = list(ExposureTypeOptions())
    for exposure_type in exposure_types:
        if exposure_type not in options:
            logger.print_and_log_error(u"'{0}' is not a valid exposure type.".format(exposure_type))
            exit(1)


def _create_file_event_filters(args):
    filters = []
    event_timestamp_filter = create_time_range_filter(EventTimestamp, args.begin, args.end)
    not event_timestamp_filter or filters.append(event_timestamp_filter)
    not args.c42_username or filters.append(DeviceUsername.is_in(args.c42_username))
    not args.actor or filters.append(Actor.is_in(args.actor))
    not args.md5 or filters.append(MD5.is_in(args.md5))
    not args.sha256 or filters.append(SHA256.is_in(args.sha256))
    not args.source or filters.append(Source.is_in(args.source))
    not args.file_name or filters.append(FileName.is_in(args.file_name))
    not args.file_path or filters.append(FilePath.is_in(args.file_path))
    not args.process_owner or filters.append(ProcessOwner.is_in(args.process_owner))
    not args.tab_url or filters.append(TabURL.is_in(args.tab_url))
    _try_append_exposure_types_filter(filters, args.include_non_exposure, args.type)
    return filters


def _try_append_exposure_types_filter(filters, include_non_exposure_events, exposure_types):
    _exposure_filter = _create_exposure_type_filter(include_non_exposure_events, exposure_types)
    if _exposure_filter:
        filters.append(_exposure_filter)


def _create_exposure_type_filter(include_non_exposure_events, exposure_types):
    if include_non_exposure_events and exposure_types:
        logger.print_and_log_error(u"Cannot use exposure types with `--include-non-exposure`.")
        exit(1)
    if exposure_types:
        return ExposureType.is_in(exposure_types)
    if not include_non_exposure_events:
        return ExposureType.exists()



# -*- coding: utf-8 -*-

from code42cli.logger import get_progress_logger


class ProgressBar(object):
    _FILL = u"█"
    _LENGTH = 100

    def __init__(self, total_items, logger=None):
        self._total_items = total_items
        self._logger = logger or get_progress_logger()

    def update(self, iteration, message):
        bar = self._create_bar(iteration)
        progress = u"{}  {}".format(bar, message.strip())
        self._logger.info(progress)

    def _create_bar(self, iteration):
        fill_length = self._calculate_fill_length(iteration)
        return self._FILL * fill_length + u"-" * (self._LENGTH - fill_length)

    def _calculate_fill_length(self, idx):
        filled_length = int(self._LENGTH * idx // self._total_items)
        return filled_length

    def clear_bar_and_print_final(self, final_message):
        clear = (self._LENGTH + len(final_message)) * u" "
        self._logger.info(u"{}{}\n".format(final_message, clear))



from code42cli.compat import range

_BATCH_SIZE = 500


def get_alert_details(sdk, alert_summary_list):
    alert_ids = [alert[u"id"] for alert in alert_summary_list]
    batches = [alert_ids[i : i + _BATCH_SIZE] for i in range(0, len(alert_ids), _BATCH_SIZE)]
    results = []
    for batch in batches:
        r = sdk.alerts.get_details(batch)
        results.extend(r[u"alerts"])
    results = sorted(results, key=lambda x: x[u"createdAt"], reverse=True)
    return results
